# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2006 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from ..support import decorators

import inspect

class NTInfo (dict):
    """
    A container for information attached to a notification.
    This class is a dictionary-like object used:

    1. As class when defining notification methods in observers, as it
       contains the flags identifying the notification types.

    2. As class instance as parameter when a notification methods is
       called in observers.


    **Notification Type Flags**
    
    Notification methods are declared either statically or dynamically 
    through :meth:`Observer.observe`. In both cases the type of the
    notification is defined by setting to `True` some flags. Flags can
    be set in any combination for multi-type notification
    methods. Flags are:

    assign
       For notifications issued when OPs are assigned.
    before
       For notifications called before a modifying method is called.
    after
       For notifications called after a modifying method is called.
    signal
       For notifications called when a signal is emitted.

    
    **Instance content**

    Instances of class `NTInfo` will be received as the last argument
    (`info`) of any notification method::

      def notification_method(self, model, name, info)

    NTInfo is a dictionary (with some particular behaviour added)
    containing some information which is independent on the
    notification type, and some other information wich depends on the
    notification type.


    **Common to all types**

    For all notification types, NTInfo contains:

    model
       the model containing the observable property triggering the
       notification. `model` is also passed as first argument of the
       notification method.

    prop_name
       the name of the observable property triggering the notification. `name`
       is also passed as second argument of the notification method.
      
    Furthermore, any keyword argument not listed here is copied
    without modification into `info`.

    There are further information depending on the specific
    notification type:

    **For Assign-type**

    assign
       flag set to `True`

    old
       the value that the observable property had before being
       changed.

    new
       the new value that the observable property has been
       changed to.


    **For Before method call type**

    before
       flag set to `True`

    instance
       the object instance which the method that is being called belongs to.

    method_name
       the name of the method that is being called. 

    args
       tuple of the arguments of the method that is being called. 

    kwargs
       dictionary of the keyword arguments of the method that
       is being called.


    **For After method call type**

    after
       flag set to `True`

    instance
       the object instance which the method that has been 
       called belongs to.

    method_name
       the name of the method that has been called. 

    args
       tuple of the arguments of the method that has been called. 

    kwargs
       dictionary of the keyword arguments of the method that
       has been called.

    result
       the value returned by the method which has been called. 

    **For Signal-type**

    signal
       flag set to `True`

    arg
       the argument which was optionally specified when invoking
       emit() on the signal observable property.

    **Information access**

    The information carried by a NTInfo instance passed to a
    notification method can be retrieved using the instance as a
    dictionary, or accessing directly to the information as an
    attribute of the instance. For example::
       
       # This is a multi-type notification
       @Observer.observe("op1", assign=True, hello="Ciao")
       @Observer.observe("op2", after=True, before=True)
       def notify_me(self, model, name, info):
           assert info["model"] == model # access as dict key
           assert info.prop_name == name # access as attribute

           if "assign" in info:
              assert info.old == info["old"]
              assert "hello" in info and "ciao" == info.hello
              print "Assign from", info.old, "to", info.new
           else:
              assert "before" in info or "after" in info
              assert "hello" not in info
              print "Method name=", info.method_name
              if "after" in info: print "Method returned", info.result    
              pass
              
           return   

    As already told, the type carried by a NTInfo instance can be
    accessed through boolean flags `assign`, `before`, `after` and
    `signal`. Furthermore, any other information specified at
    declaration time (keyword argument 'hello' in the previous
    example) will be accessible in the corresponding notification
    method.

    .. versionadded:: 1.99.1

    """

    # At least one of the keys in this set is required when constructing
    __ONE_REQUESTED = frozenset("assign before after signal".split())
    __ALL_REQUESTED = frozenset("model prop_name".split())

    def __init__(self, _type, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # checks the content provided by the user
        if not (_type in self and self[_type]):
            raise KeyError("flag '%s' must be set in given arguments" % _type)

        # all requested are provided by the framework, not the user
        assert NTInfo.__ALL_REQUESTED <= set(self)

        # now removes all type-flags not related to _type
        for flag in NTInfo.__ONE_REQUESTED:
            if flag != _type and flag in self: del self[flag]
            pass

        return

    def __getattr__(self, name):
        """
        All dictionary keys are also available as attributes.
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError("NTInfo object has no attribute '%s'.\n"
                                 "Existing attributes are: %s" % (name, str(self)))
        pass

    pass # end of class NTInfo
# ----------------------------------------------------------------------


class Observer (object):
    """
    .. note::

       Most methods in this class are used internally by the
       framework.  Do not override them in subclasses.
    """

    # this is internal
    _CUST_OBS_ = "__custom_observes__"
    # ----------------------------------------------------------------------


    @classmethod
    @decorators.good_decorator_accepting_args
    def observe(cls, *args, **kwargs):
        """
        Mark a method as recieving notifications. Comes in two flavours:

        .. method:: observe(name, **types)
           :noindex:

           A decorator living in the class. Can be applied more than once to
           the same method, provided the names differ.
           
           *name* is the property we want to be notified about as a string.
           
           *types* are boolean values denoting the types of
           notifications desired. At least one of the following has to be
           passed as True: assign, before, after, signal.

           Excess keyword arguments are passed to the method as part of the
           info dictionary.

        .. method:: observe(callable, name, **types)
           :noindex:

           An instance method to define notifications at runtime. Works as
           above.
           
           *callable* is the method to send notifications to. The effect will
           be as if this had been decorated.

        In all cases the notification method must take exactly three
        arguments: the model object, the name of the property that changed,
        and an :class:`NTInfo` object describing the change.

        .. warning::
      
           Due to limitation in the dynamic registration (in version
           1.99.1), declarations of dynamic notifications must occur
           before registering self as an observer of the models whose
           properties the notifications are supposed to be
           observing. A hack for this limitation, is to first relieve
           any interesting model before dynamically register the
           notifications, and then re-observe those models.

        .. versionadded:: 1.99.1
        """

        @decorators.good_decorator
        def _decorator(_notified):
            # marks the method with observed properties
            _list = getattr(_notified, Observer._CUST_OBS_, list())
            _list.append((name, kwargs))
            setattr(_notified, Observer._CUST_OBS_, _list)
            return _notified

        # handles arguments
        if args and isinstance(args[0], cls):
            # used as instance method, for declaring notifications
            # dynamically
            if len(args) != 3:
                raise TypeError("observe() takes exactly three arguments"
                                " when called (%d given)" % len(args))

            self = args[0]
            notified = args[1]
            name = args[2]

            assert isinstance(self, Observer), "Method Observer.observe " \
                "must be called with an Observer instance as first argument"
            if not callable(notified):
                raise TypeError("Second argument of observe() must be a callable")
            if type(name) != str:
                raise TypeError("Third argument of observe() must be a string")

            self.__register_notification(name, notified, kwargs)
            return None

        # used statically as decorator
        if len(args) != 1:
            raise TypeError("observe() takes exactly one argument when used"
                            " as decorator (%d given)" % len(args))
        name = args[0]
        if type(name) != str:
            raise TypeError("First argument of observe() must be a string")
        return _decorator
    # ----------------------------------------------------------------------


    def __init__(self, *args, **kwargs):
        """
        *model* is passed to :meth:`observe_model` if given.
        
        *spurious* indicates interest to be notified even when
        the value hasn't changed, like for: ::

         model.prop = model.prop

        .. versionadded:: 1.2.0
           Before that observers had to filter out spurious
           notifications themselves, as if the default was `True`. With
           :class:`~mvc.observable.Signal` support this is no longer
           necessary.
        """
        model = kwargs.pop("model", None)
        spurious = kwargs.pop("spurious", False)
        super(Observer, self).__init__(*args, **kwargs)

        # --------------------------------------------------------- #
        # This turns the decorator 'observe' an instance method
        def __observe(*args, **kwargs): self.__original_observe(self, *args, **kwargs)
        __observe.__name__ = self.observe.__name__
        __observe.__doc__ = self.observe.__doc__
        self.__original_observe = self.observe
        self.observe = __observe
        # --------------------------------------------------------- #

        self.__accepts_spurious__ = spurious

        # NOTE: In rev. 202 these maps were unified into
        #   __CUST_OBS_MAP only (the map contained pairs (method,
        #   args). However, this broke backward compatibility of code
        #   accessing the map through
        #   get_observing_methods. Now the informatio is split
        #   and the original information restored. To access the
        #   additional information (number of additional arguments
        #   required by observing methods) use the newly added methods.

        # Private maps: do not change/access them directly, use
        # methods to access them:
        self.__CUST_OBS_MAP = {} # prop name --> set of observing methods
        self.__CUST_OBS_KWARGS = {} # observing method --> flag

        processed_props = set() # tracks already processed properties

        # searches all custom observer methods
        for cls in inspect.getmro(type(self)):
            # list of (method-name, method-object, list of (prop-name, kwargs))
            meths = [ (name, meth, getattr(meth, Observer._CUST_OBS_))
                      for name, meth in cls.__dict__.items()
                      if (inspect.isfunction(meth) and
                          hasattr(meth, Observer._CUST_OBS_)) ]

            # props processed in this class. This is used to avoid
            # processing the same props in base classes.
            cls_processed_props = set()

            # since this is traversed top-bottom in the mro, the
            # first found match is the one to care
            for name, meth, pnames_ka in meths:
                _method = getattr(self, name) # the most top avail method

                # WARNING! Here we store the top-level method in the
                # mro, not the (unbound) method which has been
                # declared by the user with the decorator.
                for pname, ka in pnames_ka:
                    if pname not in processed_props:
                        self.__register_notification(pname, _method, ka)
                        cls_processed_props.add(pname)
                        pass
                    pass
                pass

            # accumulates props processed in this class
            processed_props |= cls_processed_props
            pass # end of loop over classes in the mro

        if model is not None:
            self.observe_model(model)
        return

    def observe_model(self, model):
        """Starts observing the given model"""
        return model.register_observer(self)

    def relieve_model(self, model):
        """Stops observing the given model"""
        return model.unregister_observer(self)

    def accepts_spurious_change(self):
        """
        Returns True if this observer is interested in receiving
        spurious value changes. This is queried by the model when
        notifying a value change."""
        return self.__accepts_spurious__

    def get_observing_methods(self, prop_name):
        """
        Return a possibly empty set of callables registered with
        :meth:`observe` for *prop_name*.
        
        .. versionadded:: 1.99.1
           Replaces :meth:`get_custom_observing_methods`.
        """
        return self.__CUST_OBS_MAP.get(prop_name, set())

    # this is done to keep backward compatibility
    get_custom_observing_methods = get_observing_methods


    def get_observing_method_kwargs(self, prop_name, method):
        """
        Returns the keyword arguments which were specified when
        declaring a notification method, either statically of
        synamically with :meth:`Observer.observe`.

        *method* a callable that was registered with
        :meth:`observes`.
        
        :rtype: dict
        """
        return self.__CUST_OBS_KWARGS[(prop_name, method)]


    def remove_observing_method(self, prop_names, method):
        """
        Remove dynamic notifications.
        
        *method* a callable that was registered with :meth:`observe`.
        
        *prop_names* a sequence of strings. This need not correspond to any
        one `add` call.

        .. note::

           This can revert the effects of a decorator at runtime. Don't.
        """
        for prop_name in prop_names:
            _set = self.__CUST_OBS_MAP.get(prop_name, set())
            if method in _set: _set.remove(method)
            key = (prop_name, method)
            if key in self.__CUST_OBS_KWARGS: del self.__CUST_OBS_KWARGS[key]
            pass

        return

    def is_observing_method(self, prop_name, method):
        """
        Returns `True` if the given method was previously added as an
        observing method, either dynamically or via decorator.
        """
        return (prop_name, method) in self.__CUST_OBS_KWARGS


    def __register_notification(self, prop_name, method, kwargs):
        """Internal service which associates the given property name
        to the method, and the (prop_name, method) with the given
        kwargs dictionary. If needed merges the dictionary, if the
        given (prop_name, method) pair was already registered (in this
        case the last registration wins in case of overlapping.)

        If given prop_name and method have been already registered, a
        ValueError exception is raised."""

        key = (prop_name, method)
        if key in self.__CUST_OBS_KWARGS:
            raise ValueError("In %s method '%s' has been declared "
                             "to be a notification for property '%s' "
                             "multiple times (only one is allowed)." % \
                                 (self.__class__,
                                  method.__name__, prop_name))

        # fills the internal structures
        if prop_name not in self.__CUST_OBS_MAP:
            self.__CUST_OBS_MAP[prop_name] = set()
            pass
        self.__CUST_OBS_MAP[prop_name].add(method)

        self.__CUST_OBS_KWARGS[key] = kwargs
        return

    pass # end of class
# ----------------------------------------------------------------------



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from .utils import run_dialog

class MessageDialog(Gtk.MessageDialog):

    accept_responses = (
        Gtk.ResponseType.ACCEPT, # @UndefinedVariable
        Gtk.ResponseType.YES, # @UndefinedVariable
        Gtk.ResponseType.APPLY, # @UndefinedVariable
        Gtk.ResponseType.OK # @UndefinedVariable
    )

    def __init__(self,
             message, parent=None,
             type=Gtk.MessageType.INFO,  # @ReservedAssignment
             flags=Gtk.DialogFlags.DESTROY_WITH_PARENT,
             buttons=Gtk.ButtonsType.NONE,
             persist=False,
             title=None):
        super(MessageDialog, self).__init__(
            parent=parent,
            type=type,
            flags=flags,
            buttons=buttons)
        self.persist = persist
        self.set_markup(message)
        if title is not None:
            self.set_title(title)

    #override
    def run(self, *args, **kwargs):
        kwargs['destroy'] = not self.persist
        return run_dialog(self, *args, **kwargs)

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Pango

from pyxrd.generic.views.treeview_tools import (
    new_text_column,
    new_pb_column,
    new_combo_column,
    create_float_data_func,
    setup_treeview
)
from pyxrd.generic.views.combobox_tools import add_combo_text_column
from pyxrd.generic.views import InlineObjectListStoreView
from pyxrd.generic.controllers import (
    DialogController,
    InlineObjectListStoreController,
    BaseController,
)

from pyxrd.phases.views import EditAtomRatioView, EditAtomContentsView
from pyxrd.phases.models.atom_relations import AtomRelation, AtomRatio, AtomContents, AtomContentObject

class AtomComboMixin(object):

    extra_props = []
    custom_handler_names = []

    def reset_combo_box(self, name):
        if self.model.component is not None:
            # Get store, reset combo
            store = self.model.create_prop_store(self.extra_props)
            combo = self.view[self.view.widget_format % name]
            combo.clear()
            combo.set_model(store)

            # Add text column:
            def get_name(layout, cell, model, itr, data=None):
                obj, lbl = model.get(itr, 0, 2)
                if callable(lbl): lbl = lbl(obj)
                cell.set_property("markup", lbl)
            add_combo_text_column(combo, data_func=get_name)

            # Set the selected item to active:
            prop = getattr(self.model, name)
            if prop is not None:
                prop = tuple(prop)
                for row in store:
                    if tuple(store.get(row.iter, 0, 1)) == prop:
                        combo.set_active_iter(row.iter)
                        break

            return combo, store
        else:
            return None, None

    @staticmethod
    def custom_handler(controller, prop, prefix):
        if prop.label in controller.custom_handler_names:
            combo, store = controller.reset_combo_box(prop.label) # @UnusedVariable

            if combo is not None and store is not None:
                def on_changed(combo, user_data=None):
                    itr = combo.get_active_iter()
                    if itr is not None:
                        val = combo.get_model().get(itr, 0, 1)
                        setattr(controller.model, getattr(combo, 'model_prop'), val)
                setattr(combo, 'model_prop', prop.label)
                combo.connect('changed', on_changed)

                def on_item_changed(*args):
                    controller.reset_combo_box(prop.label)

                if controller.is_observing_method("atoms_changed", on_item_changed):
                    controller.remove_observing_method("atoms_changed", on_item_changed)
                controller.observe(on_item_changed, "atoms_changed", signal=True)

        else: return False
        return True

    pass # end of class

class EditUnitCellPropertyController(BaseController, AtomComboMixin):
    """ 
        Controller for the UnitCellProperty models (a and b cell lengths)
    """

    custom_handler_names = ["prop", ]
    widget_handlers = {
        'combo': 'custom_handler',
    }

    def __init__(self, extra_props, **kwargs):
        super(EditUnitCellPropertyController, self).__init__(**kwargs)
        self.extra_props = extra_props

    def register_adapters(self):
        BaseController.register_adapters(self)
        self.update_sensitivities()

    def update_sensitivities(self):
        self.view['ucp_value'].set_sensitive(not self.model.enabled)
        self.view['box_enabled'].set_sensitive(self.model.enabled)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @BaseController.observe("enabled", assign=True)
    def notif_enabled_changed(self, model, prop_name, info):
        self.update_sensitivities()

    pass # end of class

class EditAtomRatioController(DialogController, AtomComboMixin):
    """ 
        Controller for the atom ratio edit dialog
    """
    custom_handler_names = ["atom1", "atom2"]
    widget_handlers = {
        'custom': 'custom_handler',
    }

    pass # end of class

class EditAtomContentsController(DialogController):
    """ 
        Controller for the atom contents edit dialog
    """
    
    auto_adapt_excluded = [
        "atom_contents",
    ]

    contents_list_view = None
    contents_list_controller = None

    def __init__(self, *args, **kwargs):
        super(EditAtomContentsController, self).__init__(*args, **kwargs)
        # Create atom contents controller:
        self.contents_list_view = InlineObjectListStoreView(parent=self.view)
        self.contents_list_controller = ContentsListController("atom_contents", model=self.model, view=self.contents_list_view, parent=self)
        # Set subview:
        self.view.set_contents_list_view(self.contents_list_view.get_top_widget())

    pass # end of class

class ContentsListController(InlineObjectListStoreController):
    """ 
        Controller for the atom contents ListStore
    """
    new_val = None
    auto_adapt = False # FIXME

    treemodel_class_type = AtomContentObject

    def _reset_treeview(self, tv, model):
        setup_treeview(tv, model, sel_mode='MULTIPLE', reset=True)
        tv.set_model(model)

        # Atom column:
        self.combo_model = self.model.create_prop_store()
        self.combo_model2 = Gtk.ListStore(str)
        for row in self.combo_model:
            self.combo_model2.append(row[2:3])
              
        def atom_renderer(column, cell, model, itr, *args):
            obj = model.get_value(itr, 0)
            if hasattr(obj, "name"):
                cell.set_property('text', obj.name)
            else:
                cell.set_property('text', '#NA#')      
        
        tv.append_column(new_combo_column(
            "Atoms",
            data_func=atom_renderer,
            changed_callback=self.on_atom_changed,
            edited_callback=self.on_atom_edited,
            xalign=0.0,
            expand=False,
            has_entry=False,
            model=self.combo_model2,
            text_column=0,
            editable=True))

        # Content column:
        def on_float_edited(rend, path, new_text, col):
            itr = model.get_iter(path)
            try:
                model.set_value(itr, col, float(new_text))
            except ValueError:
                logger.exception("Invalid value entered ('%s')!" % new_text)
            return True
        tv.append_column(new_text_column('Default contents', text_col=2, xalign=0.0,
            editable=True,
            data_func=create_float_data_func(),
            edited_callback=(on_float_edited, (2,))))

    def _setup_treeview(self, tv, model):
        self._reset_treeview(tv, model)

    def __init__(self, treemodel_property_name, **kwargs):
        super(ContentsListController, self).__init__(
            treemodel_property_name=treemodel_property_name,
            enable_import=False, enable_export=False, **kwargs
        )

    def create_new_object_proxy(self):
        return AtomContentObject(None, None, 1.0)

    def on_atom_changed(self, combo, path, new_iter):
        # translate dummy iter to real iter:
        new_iter = self.combo_model.get_iter(self.combo_model2.get_path(new_iter))
        self.new_val = self.combo_model.get(new_iter, 0, 1)
        pass

    def on_atom_edited(self, combo, path, new_text, model=None):
        if self.new_val:
            new_atom, new_prop = self.new_val
            self.model.set_atom_content_values(path, new_atom, new_prop)
            self.new_val = None
        return True


    pass # end of class

class EditAtomRelationsController(InlineObjectListStoreController):
    """ 
        Controller for the components' atom relations ObjectListStore
    """
    file_filters = AtomRelation.Meta.file_filters
    auto_adapt = False
    treemodel_class_type = AtomRelation

    add_types = [
        ("Ratio", AtomRatio, EditAtomRatioView, EditAtomRatioController),
        ("Contents", AtomContents, EditAtomContentsView, EditAtomContentsController),
    ]

    def _reset_treeview(self, tv, model):
        setup_treeview(tv, model, sel_mode='MULTIPLE', reset=True)
        tv.set_model(model)

        # Name column:
        def text_renderer(column, cell, model, itr, args=None):
            driven_by_other = model.get_value(itr, model.c_driven_by_other)
            cell.set_property('editable', not driven_by_other)
            cell.set_property('style', Pango.Style.ITALIC if driven_by_other else Pango.Style.NORMAL)
        col = new_text_column(
            'Name',
            data_func=text_renderer,
            editable=True,
            edited_callback=(self.on_item_cell_edited, (model, model.c_name)),
            resizable=False,
            text_col=model.c_name)
        setattr(col, "col_descr", 'Name')
        tv.append_column(col)

        # Value of the relation:
        float_rend = create_float_data_func()
        def data_renderer(column, cell, model, itr, args=None):
            text_renderer(column, cell, model, itr, args)
            float_rend(column, cell, model, itr, args)
        col = new_text_column(
            'Value',
            data_func=data_renderer,
            editable=True,
            edited_callback=(self.on_item_cell_edited, (model, model.c_value)),
            resizable=False,
            text_col=model.c_value)
        setattr(col, "col_descr", 'Value')
        tv.append_column(col)

        # Up, down and edit arrows:
        def setup_image_button(image, colnr):
            col = new_pb_column(" ", resizable=False, expand=False, stock_id=image)
            setattr(col, "col_descr", colnr)
            tv.append_column(col)
        setup_image_button("213-up-arrow", "Up")
        setup_image_button("212-down-arrow", "Down")
        setup_image_button("030-pencil", "Edit")

    def _setup_treeview(self, tv, model):
        tv.connect('button-press-event', self.tv_button_press)
        self._reset_treeview(tv, model)

    def __init__(self, **kwargs):
        super(EditAtomRelationsController, self).__init__(
            enable_import=False, enable_export=False, **kwargs)

    def create_new_object_proxy(self):
        return self.add_type(parent=self.model)

    def tv_button_press(self, tv, event):
        relation = None
        ret = tv.get_path_at_pos(int(event.x), int(event.y))

        if ret is not None:
            path, col, x, y = ret
            model = tv.get_model()
            relation = model.get_user_data_from_path(path)
            column = getattr(col, "col_descr")
        if event.button == 1 and relation is not None:
            column = getattr(col, "col_descr")
            if column == "Edit":
                self._edit_item(relation)
                return True
            elif column == "Up":
                self.model.move_atom_relation_up(relation)
                return True
            elif column == "Down":
                self.model.move_atom_relation_down(relation)
                return True


    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

renderer_map = {
    'text': Gtk.CellRendererText,
    'accel': Gtk.CellRendererAccel,
    'combo': Gtk.CellRendererCombo,
    'spin': Gtk.CellRendererSpin,
    'pixbuf': Gtk.CellRendererPixbuf,
    'progress': Gtk.CellRendererProgress,
    'spinner': Gtk.CellRendererSpinner,
    'toggle': Gtk.CellRendererToggle,
}

def get_default_renderer(type, **kwargs):
    """
        Creates a CellRendere of type 'type' and sets any attributes passed with
        the key-word arguments. Underscores in variable names are replaced with
        dashes in the proces.
    """
    rend = renderer_map.get(type, type)()
    for key, val in kwargs.items():
        rend.set_property(key.replace("_", "-"), val)
    return rend
    
def parse_callback(callback, reduce=True):
    """
        Parses callbacks for CellRenderers: it splits the 
        callback from its arguments if present. Additionally this method will 
        not create singleton argument lists, but pass them as a single argument.
        
        Returns the callback and its argument(s) (or an empty tuple)
    """
    args = tuple()
    try:
        callback, args = callback
    except TypeError as ValueError:
        pass
    #deconvolve things:
    if reduce and len(args) == 1: args = args[0]
    return callback, args
    
def parse_kwargs(**kwargs):
    """
        Parses key-word arguments.
        It checks for the presence of key-words ending with '_col', these are
        popped and stored in a seperate dictionary, as they are to be passed
        to the constructor of the actual column or combobox (attribute mappings)
        In addition it sets a number of default attributes for the CellRenderer.
        
        Returns a tuple containing a dict with the CellRenderer attributes and a
        dict with the TreeViewColumn attribute mappings
    """
    kwargs["xalign"] = kwargs.get("xalign", 0.5)
    kwargs["yalign"] = kwargs.get("yalign", 0.5)
    
    col_attrs = dict()
    for key, value in dict(kwargs).items():
        if key.endswith("_col"):
            col_attrs[key[:-4]] = value
            kwargs.pop(key)
    return kwargs, col_attrs



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
from pkg_resources import resource_filename # @UnresolvedImport

from pyxrd.generic.exceptions import AlreadyRegistered, NotRegistered

class DataRegistry(object):
    """
        Class for registering data directories and files so they're
        not hard-coded everywhere.
    """

    __data_directories = None
    __data_files = None
    _base_dir = None

    def __init__(self, dirs=[], files=[], *args, **kwargs):
        super(DataRegistry, self).__init__(*args, **kwargs)
        self.__data_directories = {}
        self.__data_files = {}
        for name, path, parent in dirs:
            self.register_data_directory(name, path, parent=parent)

        for name, path, parent in files:
            self.register_data_file(name, path, parent=parent)

    def __parse_parent(self, path, parent=None):
        # Adds parent path:
        if parent is not None:
            if parent is not None:
                try:
                    path = os.path.join(self.__data_directories[parent], path)
                except KeyError:
                    raise NotRegistered("The data directory named '%s' was not found in the registry" % parent)
        elif path.startswith("./"):
            path = resource_filename("pyxrd.data", path)
        return path

    def register_data_file(self, name, path, parent=None):
        """
            Registers a data file at 'path' called 'name'.
            If this file is inside a registered data directory, you
            can use a relative path by setting the 'parent' keyword argument
            to the name of the parent data-directory. The parent path
            will then be appended to the file's path. Paths are considered to be
            relative to data package.
            
            Note: names are always made full-caps!
            
            If you try to re-register an existing data file, an AlreadyRegistered
            exception will be raised. Similarly if you pass in an unregistered
            parent directory name, NotRegistered will be raised.
        """
        name = name.upper()
        if not name in self.__data_files:
            path = self.__parse_parent(path, parent=parent)
            self.__data_files[name] = path
        else:
            raise AlreadyRegistered("the data file named '%s' was already registered" % name)

    def register_data_directory(self, name, path, parent=None):
        """
            Registers a data directory at 'path' called 'name'.
            If this is a sub-directory of another registered data directory, you
            can use a relative path by setting the 'parent' keyword argument
            to the name of the parent data-directory. The parent path
            will then be appended to the child's path. Paths are considered to be
            relative to data package.
            
            Note: names are always made full-caps!
            
            If you try to re-register an existing data directory, an AlreadyRegistered
            exception will be raised. Similarly if you pass in an unregistered
            parent directory name, NotRegistered will be raised.
        """
        name = name.upper()
        if not name in self.__data_directories:
            path = self.__parse_parent(path, parent=parent)
            self.__data_directories[name] = path
            try: # Try to create this path:
                os.makedirs(path)
            except OSError:
                pass
        else:
            raise AlreadyRegistered("The data directory named '%s' was already registered" % name)

    def get_directory_path(self, name):
        """
            Gets the absolute path to a data directory named 'name'
        """
        try:
            path = self.__data_directories[name]
            if not os.path.isdir(path):
                return path
            else:
                return path
        except KeyError:
            raise NotRegistered("The data directory named '%s' was not found in the registry" % name)

    def get_all_directories(self):
        """
            Returns a generator looping over all directories in the registry,
            excluding the project path.
        """
        for path in list(self.__data_directories.values()):
            yield path

    def get_file_path(self, name):
        """
            Gets the absolute path to a data file named 'name'
        """
        try:
            return self.__data_files[name]
        except KeyError:
            raise NotRegistered("The data file named '%s' was not found in the registry" % name)

    def get_all_files(self):
        """
            Returns a generator looping over all directories in the registry,
            excluding the project path.
        """
        for path in list(self.__data_directories.values()):
            yield resource_filename("pyxrd.data", path)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from mvc.models.properties import (
    LabeledProperty, StringProperty, BoolProperty,
    ReadOnlyMixin
)

from pyxrd.generic.models.base import ChildModel

from .mixins import _RefinementBase, RefinementValue, RefinementGroup

class RefinableWrapper(ChildModel):
    """
        Wrapper class for refinables easing the retrieval of certain
        properties for the different types of refinables.
        Can be used with an ObjectTreeStore.
    """

    # MODEL INTEL:
    class Meta(ChildModel.Meta):
        parent_alias = "mixture"

    # PROPERTIES:

    #: The wrapped object
    obj = LabeledProperty(
        default=None, text="Wrapped object", tabular=True)

    #: The property descriptor object for the attribute
    prop_descr = LabeledProperty(
        default=None, text="Property descriptor", tabular=True)

    #: The Property label:
    @StringProperty(
        default="", text="Property label", tabular=True, mix_with=(ReadOnlyMixin,))
    def label(self):
        return self.prop_descr.label

    #: A flag indicating whether this is wrapper is representing the group
    #: (True) or a member of the group (False):
    is_grouper = BoolProperty(
        default=False, text="Is grouper", tabular=True, mix_with=(ReadOnlyMixin,))

    #: The inherit attribute name:
    @LabeledProperty(
        default=None, text="Inherit from label", mix_with=(ReadOnlyMixin,))
    def inherit_from(self):
        return self.prop_descr.inherit_from if self.prop_descr else None

    #: The (possibly mathtext) label for the refinable property:
    @StringProperty(
        default="", text="Title", tabular=True, mix_with=(ReadOnlyMixin,))
    def title(self):
        if (isinstance(self.obj, RefinementGroup) and self.is_grouper) or isinstance(self.obj, RefinementValue):
            return self.obj.refine_title
        else:
            if getattr(self.prop_descr, "math_text", None) is not None:
                return self.prop_descr.math_text
            else:
                return self.prop_descr.text

    #: The (pure text) label for the refinable property:
    @StringProperty(
        default="", text="Text title", tabular=True, mix_with=(ReadOnlyMixin,))
    def text_title(self):
        if (isinstance(self.obj, RefinementGroup) and self.is_grouper) or isinstance(self.obj, RefinementValue):
            return self.obj.refine_title
        else:
            return self.prop_descr.text

    @StringProperty(
        default="", text="Descriptor", tabular=True, mix_with=(ReadOnlyMixin,))
    def text_descriptor(self):
        """ Return a longer title that also describes this property's relations """

        # This gets the phase and/or component name for the group or value:
        data = self.obj.refine_descriptor_data

        # Here we still need to get the actual property title:
        data["property_name"] = self.text_title

        return "%(phase_name)s | %(component_name)s | %(property_name)s" % data

    #: The actual value of the refinable property:
    @LabeledProperty(
        default=None, text="Value", tabular=True)
    def value(self):
        if isinstance(self.obj, RefinementValue):
            return self.obj.refine_value
        elif not self.is_grouper:
            return getattr(self.obj, self.label)
        else:
            return ""
    @value.setter
    def value(self, value):
        value = max(min(value, self.value_max), self.value_min)
        if self.is_grouper:
            raise AttributeError("Cannot set the value for a grouping RefinableWrapper")
        elif isinstance(self.obj, RefinementValue):
            self.obj.refine_value = value
        else:
            setattr(self.obj, self.label, value)

    #: Whether or not this property is inherited from another object
    @BoolProperty(
        default=False, text="Inherited", tabular=True, mix_with=(ReadOnlyMixin,))
    def inherited(self):
        return self.inherit_from is not None and hasattr(self.obj, self.inherit_from) and getattr(self.obj, self.inherit_from)

    #: Whether or not this property is actually refinable
    @BoolProperty(
        default=False, text="Refinable", tabular=True, mix_with=(ReadOnlyMixin,))
    def refinable(self):
        if isinstance(self.obj, _RefinementBase):
            # We have a _RefinementBase property (group or value)
            if isinstance(self.obj, RefinementGroup):
                if self.is_grouper: # the grouper itself
                    return False
                else: # attribute of the grouper
                    return (not self.inherited) and self.obj.children_refinable
            elif isinstance(self.obj, RefinementValue):
                return (not self.inherited) and self.obj.is_refinable
        else:
            # This is actually impossible, but what the hack...
            return (not self.inherited)

    #: The refinement info object for the refinable property
    @LabeledProperty(
        default=None, text="Refinement info", tabular=True, mix_with=(ReadOnlyMixin,))
    def ref_info(self):
        if (isinstance(self.obj, RefinementGroup) and self.is_grouper) or isinstance(self.obj, RefinementValue):
            return self.obj.refine_info
        else:
            name = self.prop_descr.get_refinement_info_name()
            if name is not None:
                ref_info = getattr(self.obj, name)
                return ref_info
            else:
                raise AttributeError("Cannot find refine info model for attribute '%s' on '%s'" % (self.label, self.obj))

    #: The minimum value for the refinable property
    @LabeledProperty(
        default=None, text="Minimum value", tabular=True)
    def value_min(self):
        return self.ref_info.minimum if self.ref_info else None
    @value_min.setter
    def value_min(self, value):
        if self.ref_info:
            self.ref_info.minimum = value

    #: The maximum value of the refinable property
    @LabeledProperty(
        default=None, text="Maximum value", tabular=True)
    def value_max(self):
        return self.ref_info.maximum if self.ref_info else None
    @value_max.setter
    def value_max(self, value):
        if self.ref_info:
            self.ref_info.maximum = value

    #: Wether this property is selected for refinement
    @BoolProperty(
        default=False, text="Refine", tabular=True)
    def refine(self):
        return self.ref_info.refine if self.ref_info else False
    @refine.setter
    def refine(self, value):
        if self.ref_info:
            self.ref_info.refine = value and self.refinable

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for a RefinableWrapper are:
                obj: the object we are wrapping a parameter for
                prop or prop_descr: the property descriptor
                is_grouper: whether or not this is a grouper object
        """
        my_kwargs = self.pop_kwargs(kwargs, "obj", "prop", "prop_descr", "is_grouper")
        super(RefinableWrapper, self).__init__(**kwargs)
        kwargs = my_kwargs

        self.obj = self.get_kwarg(kwargs, None, "obj")
        self.prop_descr = self.get_kwarg(kwargs, None, "prop_descr", "prop")
        self._is_grouper = self.get_kwarg(kwargs, False, "is_grouper")

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from math import pi, log

import numpy as np

from mvc.observers import ListObserver
from mvc.models.properties import (
    StringProperty, SignalMixin, ReadOnlyMixin, FloatProperty, LabeledProperty,
    ObserveMixin, ListProperty, BoolProperty
)

from pyxrd.data import settings

from pyxrd.generic.io import storables, Storable
from pyxrd.generic.models import ExperimentalLine, CalculatedLine, DataModel
from pyxrd.generic.utils import not_none
from pyxrd.generic.models.lines import PyXRDLine

from pyxrd.calculations.peak_detection import peakdetect
from pyxrd.calculations.data_objects import SpecimenData

from pyxrd.goniometer.models import Goniometer

from pyxrd.file_parsers.xrd_parsers import xrd_parsers
from pyxrd.file_parsers.exc_parsers import exc_parsers

from .markers import Marker
from .statistics import Statistics


@storables.register()
class Specimen(DataModel, Storable):
    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Specimen"

        export_filters = xrd_parsers.get_export_file_filters()
        excl_filters = exc_parsers.get_import_file_filters()

    _data_object = None
    @property
    def data_object(self):
        self._data_object.goniometer = self.goniometer.data_object
        self._data_object.range_theta = self.__get_range_theta()
        self._data_object.selected_range = self.get_exclusion_selector()
        self._data_object.z_list = self.get_z_list()
        try:
            self._data_object.observed_intensity = np.copy(self.experimental_pattern.data_y)
        except IndexError:
            self._data_object.observed_intensity = np.array([], dtype=float)
        return self._data_object

    def get_z_list(self):
        return list(self.experimental_pattern.z_data)

    project = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: The sample name
    sample_name = StringProperty(
        default="", text="Sample",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The sample name
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )


    @StringProperty(
        default="", text="Label",
        visible=False, persistent=False, tabular=True,
        mix_with=(ReadOnlyMixin,)
    )
    def label(self):
        if self.display_stats_in_lbl and (self.project is not None and self.project.layout_mode == "FULL"):
            label = self.sample_name
            label += "\nRp = %.1f%%" % not_none(self.statistics.Rp, 0.0)
            label += "\nRwp = %.1f%%" % not_none(self.statistics.Rwp, 0.0)
            return label
        else:
            return self.sample_name

    display_calculated = BoolProperty(
        default=True, text="Display calculated diffractogram",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    display_experimental = BoolProperty(
        default=True, text="Display experimental diffractogram",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    display_vshift = FloatProperty(
        default=0.0, text="Vertical shift of the plot",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed", widget_type="spin",
        mix_with=(SignalMixin,)
    )

    display_vscale = FloatProperty(
        default=0.0, text="Vertical scale of the plot",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed", widget_type="spin",
        mix_with=(SignalMixin,)
    )

    display_phases = BoolProperty(
        default=True, text="Display phases seperately",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    display_stats_in_lbl = BoolProperty(
        default=True, text="Display Rp in label",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    display_residuals = BoolProperty(
        default=True, text="Display residual patterns",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    display_residual_scale = FloatProperty(
        default=1.0, text="Residual pattern scale", minimum=0.0,
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed", widget_type="spin",
        mix_with=(SignalMixin,)
    )

    display_derivatives = BoolProperty(
        default=False, text="Display derivative patterns",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: A :class:`~pyxrd.generic.models.lines.CalculatedLine` instance
    calculated_pattern = LabeledProperty(
        default=None, text="Calculated diffractogram",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="xy_list_view",
        mix_with=(SignalMixin, ObserveMixin,)
    )

    #: A :class:`~pyxrd.generic.models.lines.ExperimentalLine` instance
    experimental_pattern = LabeledProperty(
        default=None, text="Experimental diffractogram",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="xy_list_view",
        mix_with=(SignalMixin, ObserveMixin,)
    )

    #: A list of 2-theta ranges to exclude for the calculation of the Rp factor
    exclusion_ranges = LabeledProperty(
        default=None, text="Excluded ranges",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="xy_list_view",
        mix_with=(SignalMixin, ObserveMixin)
    )

    #: A :class:`~pyxrd.goniometer.models.Goniometer` instance
    goniometer = LabeledProperty(
        default=None, text="Goniometer",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin, ObserveMixin,)
    )

    #: A :class:`~pyxrd.specimen.models.Statistics` instance
    statistics = LabeledProperty(
        default=None, text="Markers",
        visible=False, persistent=False, tabular=True,
    )

    #: A list :class:`~pyxrd.specimen.models.Marker` instances
    markers = ListProperty(
        default=None, text="Markers", data_type=Marker,
        visible=False, persistent=True, tabular=True,
        signal_name="visuals_changed", widget_type="object_list_view",
        mix_with=(SignalMixin,)
    )

    @property
    def max_display_y(self):
        """
         The maximum intensity or z-value (display y axis) 
         of the current profile (both calculated and observed)
        """
        _max = 0.0
        if self.experimental_pattern is not None:
            _max = max(_max, np.max(self.experimental_pattern.max_display_y))
        if self.calculated_pattern is not None:
            _max = max(_max, np.max(self.calculated_pattern.max_display_y))
        return _max

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for a Specimen are:
                name: the name of the specimen
                sample_name: the sample name of the specimen
                calculated_pattern: the calculated pattern
                experimental_pattern: the experimental pattern
                exclusion_ranges: the exclusion ranges XYListStore
                goniometer: the goniometer used for recording data
                markers: the specimen's markers
                display_vshift: the patterns vertical shift from its default position
                display_vscale: the patterns vertical scale (default is 1.0)
                display_calculated: whether or not to show the calculated pattern
                display_experimental: whether or not to show the experimental pattern
                display_residuals: whether or not to show the residuals
                display_derivatives: whether or not to show the 1st derivative patterns
                display_phases: whether or not to show the separate phase patterns
                display_stats_in_lbl: whether or not to display the Rp values 
                 in the pattern label
        """

        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "data_sample", "data_sample_length",
            "data_calculated_pattern", "data_experimental_pattern",
            "calc_color", "calc_lw", "inherit_calc_color", "inherit_calc_lw",
            "exp_color", "exp_lw", "inherit_exp_color", "inherit_exp_lw",
            "project_goniometer", "data_markers", "bg_shift", "abs_scale",
            "exp_cap_value", "sample_length", "absorption", "sample_z_dev",
            *[prop.label for prop in Specimen.Meta.get_local_persistent_properties()]
        )
        super(Specimen, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self._data_object = SpecimenData()

        with self.visuals_changed.hold_and_emit():
            with self.data_changed.hold_and_emit():
                self.name = self.get_kwarg(kwargs, "", "name", "data_name")
                sample_name = self.get_kwarg(kwargs, "", "sample_name", "data_sample")
                if isinstance(sample_name, bytes):
                    sample_name = sample_name.decode("utf-8", "ignore")
                self.sample_name = sample_name
                
                calc_pattern_old_kwargs = {}
                for kw in ("calc_color", "calc_lw", "inherit_calc_color", "inherit_calc_lw"):
                    if kw in kwargs:
                        calc_pattern_old_kwargs[kw.replace("calc_", "")] = kwargs.pop(kw)
                self.calculated_pattern = self.parse_init_arg(
                    self.get_kwarg(kwargs, None, "calculated_pattern", "data_calculated_pattern"),
                    CalculatedLine,
                    child=True, default_is_class=True,
                    label="Calculated Profile",
                    parent=self,
                    **calc_pattern_old_kwargs
                )

                exp_pattern_old_kwargs = {}
                for kw in ("exp_color", "exp_lw", "inherit_exp_color", "inherit_exp_lw"):
                    if kw in kwargs:
                        exp_pattern_old_kwargs[kw.replace("exp_", "")] = kwargs.pop(kw)
                self.experimental_pattern = self.parse_init_arg(
                    self.get_kwarg(kwargs, None, "experimental_pattern", "data_experimental_pattern"),
                    ExperimentalLine,
                    child=True, default_is_class=True,
                    label="Experimental Profile",
                    parent=self,
                    **exp_pattern_old_kwargs
                )

                self.exclusion_ranges = PyXRDLine(data=self.get_kwarg(kwargs, None, "exclusion_ranges"), parent=self)
                
                # Extract old kwargs if they are there:
                gonio_kwargs = {}
                sample_length = self.get_kwarg(kwargs, None, "sample_length", "data_sample_length")
                if sample_length is not None:
                    gonio_kwargs["sample_length"] = float(sample_length)
                absorption = self.get_kwarg(kwargs, None, "absorption")
                if absorption is not None: # assuming a surface density of at least 20 mg/cm²:
                    gonio_kwargs["absorption"] = float(absorption) / 0.02
                    
                # Initialize goniometer (with optional old kwargs):
                self.goniometer = self.parse_init_arg(
                    self.get_kwarg(kwargs, None, "goniometer", "project_goniometer"),
                    Goniometer, child=True, default_is_class=True,
                    parent=self, **gonio_kwargs
                )

                self.markers = self.get_list(kwargs, None, "markers", "data_markers", parent=self)
                for marker in self.markers:
                    self.observe_model(marker)
                self._specimens_observer = ListObserver(
                    self.on_marker_inserted,
                    self.on_marker_removed,
                    prop_name="markers",
                    model=self
                )

                self.display_vshift = float(self.get_kwarg(kwargs, 0.0, "display_vshift"))
                self.display_vscale = float(self.get_kwarg(kwargs, 1.0, "display_vscale"))
                self.display_calculated = bool(self.get_kwarg(kwargs, True, "display_calculated"))
                self.display_experimental = bool(self.get_kwarg(kwargs, True, "display_experimental"))
                self.display_residuals = bool(self.get_kwarg(kwargs, True, "display_residuals"))
                self.display_residual_scale = float(self.get_kwarg(kwargs, 1.0, "display_residual_scale"))
                self.display_derivatives = bool(self.get_kwarg(kwargs, False, "display_derivatives"))
                self.display_phases = bool(self.get_kwarg(kwargs, False, "display_phases"))
                self.display_stats_in_lbl = bool(self.get_kwarg(kwargs, True, "display_stats_in_lbl"))

                self.statistics = Statistics(parent=self)

                pass # end of with
            pass # end of with
        pass # end of __init__

    def __str__(self):
        return "<Specimen %s(%s)>" % (self.name, repr(self))

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @DataModel.observe("data_changed", signal=True)
    def notify_data_changed(self, model, prop_name, info):
        if model == self.calculated_pattern:
            self.visuals_changed.emit() # don't propagate this as data_changed
        else:
            self.data_changed.emit() # propagate signal

    @DataModel.observe("visuals_changed", signal=True)
    def notify_visuals_changed(self, model, prop_name, info):
        self.visuals_changed.emit() # propagate signal

    def on_marker_removed(self, item):
        with self.visuals_changed.hold_and_emit():
            self.relieve_model(item)
            item.parent = None

    def on_marker_inserted(self, item):
        with self.visuals_changed.hold_and_emit():
            self.observe_model(item)
            item.parent = self

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    @staticmethod
    def from_experimental_data(filename, parent, parser=xrd_parsers._group_parser, load_as_insitu=False):
        """
            Returns a list of new :class:`~.specimen.models.Specimen`'s loaded
            from `filename`, setting their parent to `parent` using the given
            parser. If the load_as_insitu flag is set to true, 
        """
        specimens = list()
        xrdfiles = parser.parse(filename)
        if len(xrdfiles):
            if getattr(xrdfiles[0], "relative_humidity_data", None) is not None: # we have relative humidity data
                specimen = None
                
                # Setup list variables:
                x_data = None
                y_datas = []
                rh_datas = []
                
                for xrdfile in xrdfiles:
                    # Get data we need:
                    name, sample, xy_data, rh_data = (
                        xrdfile.filename, xrdfile.name, 
                        xrdfile.data, xrdfile.relative_humidity_data
                    )
                    # Transform into numpy array for column selection
                    xy_data = np.array(xy_data)
                    rh_data = np.array(rh_data)
                    if specimen is None:
                        specimen = Specimen(parent=parent, name=name, sample_name=sample)
                        specimen.goniometer.reset_from_file(xrdfile.create_gon_file())
                        # Extract the 2-theta positions once:
                        x_data = np.copy(xy_data[:,0])
                    
                    # Add a new sub-pattern:
                    y_datas.append(np.copy(xy_data[:,1]))
                    
                    # Store the average RH for this pattern:
                    rh_datas.append(np.average(rh_data))
                    
                specimen.experimental_pattern.load_data_from_generator(zip(x_data, np.asanyarray(y_datas).transpose()), clear=True)
                specimen.experimental_pattern.y_names = ["%.1f" % f for f in rh_datas]
                specimen.experimental_pattern.z_data = rh_datas
                specimens.append(specimen)    
            else: # regular (might be multi-pattern) file
                for xrdfile in xrdfiles:
                    name, sample, generator = xrdfile.filename, xrdfile.name, xrdfile.data
                    specimen = Specimen(parent=parent, name=name, sample_name=sample)
                    # TODO FIXME:
                    specimen.experimental_pattern.load_data_from_generator(generator, clear=True)
                    specimen.goniometer.reset_from_file(xrdfile.create_gon_file())
                    specimens.append(specimen)
        return specimens

    def json_properties(self):
        props = Storable.json_properties(self)
        props["exclusion_ranges"] = self.exclusion_ranges._serialize_data()
        return props

    def get_export_meta_data(self):
        """ Returns a dictionary with common meta-data used in export functions
            for experimental or calculated data """
        return dict(
            sample=self.label + " " + self.sample_name,
            wavelength=self.goniometer.wavelength,
            radius=self.goniometer.radius,
            divergence=self.goniometer.divergence,
            soller1=self.goniometer.soller1,
            soller2=self.goniometer.soller2,
        )


    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def clear_markers(self):
        with self.visuals_changed.hold():
            for marker in list(self.markers)[::-1]:
                self.markers.remove(marker)

    def auto_add_peaks(self, tmodel):
        """
        Automagically add peak markers
        
        *tmodel* a :class:`~specimen.models.ThresholdSelector` model
        """
        threshold = tmodel.sel_threshold
        base = 1 if (tmodel.pattern == "exp") else 2
        data_x, data_y = tmodel.get_xy()
        maxtab, mintab = peakdetect(data_y, data_x, 5, threshold) # @UnusedVariable

        mpositions = [marker.position for marker in self.markers]

        with self.visuals_changed.hold():
            i = 1
            for x, y in maxtab: # @UnusedVariable
                if not x in mpositions:
                    nm = self.goniometer.get_nm_from_2t(x) if x != 0 else 0
                    new_marker = Marker(label="%%.%df" % (3 + min(int(log(nm, 10)), 0)) % nm, parent=self, position=x, base=base)
                    self.markers.append(new_marker)
                i += 1

    def get_exclusion_selector(self):
        """
        Get the numpy selector array for non-excluded data
        
        :rtype: a numpy ndarray
        """
        x = self.__get_range_theta() * 360.0 / pi # convert to degrees
        selector = np.ones(x.shape, dtype=bool)
        data = np.sort(np.asarray(self.exclusion_ranges.get_xy_data()), axis=0)
        for x0, x1 in zip(*data):
            new_selector = ((x < x0) | (x > x1))
            selector = selector & new_selector
        return selector

    def get_exclusion_xy(self):
        """
        Get an numpy array containing only non-excluded data X and Y data
                
        :rtype: a tuple containing 4 numpy ndarray's: the experimental X and Y
        data and the calculated X and Y data
        """
        ex, ey = self.experimental_pattern.get_xy_data()
        cx, cy = self.calculated_pattern.get_xy_data()
        selector = self.get_exclusion_selector(ex)
        return ex[selector], ey[selector], cx[selector], cy[selector]

    # ------------------------------------------------------------
    #      Draggable mix-in hook:
    # ------------------------------------------------------------
    def on_pattern_dragged(self, delta_y, button=1):
        if button == 1:
            self.display_vshift += delta_y
        elif button == 3:
            self.display_vscale += delta_y
        elif button == 2:
            self.project.display_plot_offset += delta_y
        pass

    def update_visuals(self, phases):
        """
            Update visual representation of phase patterns (if any)
        """
        if phases is not None:
            self.calculated_pattern.y_names = [
               phase.name if phase is not None else "" for phase in phases
            ]
            self.calculated_pattern.phase_colors = [
                phase.display_color if phase is not None else "#FF00FF" for phase in phases
            ]

    # ------------------------------------------------------------
    #      Intensity calculations:
    # ------------------------------------------------------------
    def update_pattern(self, total_intensity, phase_intensities, phases):
        """
        Update calculated patterns using the provided total and phase intensities
        """
        if len(phases) == 0:
            self.calculated_pattern.clear()
        else:
            maxZ = len(self.get_z_list())
            new_data = np.zeros((phase_intensities.shape[-1], maxZ + maxZ*len(phases))) 
            for z_index in range(maxZ):   
                # Set the total intensity for this z_index:
                new_data[:, z_index] = total_intensity[z_index]
                # Calculate phase intensity offsets:
                phase_start_index = maxZ + z_index * len(phases)
                phase_end_index = phase_start_index + len(phases)
                # Set phase intensities for this z_index:                
                new_data[:,phase_start_index:phase_end_index] = phase_intensities[:,z_index,:].transpose()
                # Store in pattern:
                self.calculated_pattern.set_data(
                    self.__get_range_theta() * 360. / pi,
                    new_data
                )
            self.update_visuals(phases)
        if settings.GUI_MODE:
            self.statistics.update_statistics(derived=self.display_derivatives)

    def convert_to_fixed(self):
        """
        Converts the experimental data from ADS to fixed slits in-place 
        (disregards the `has_ads` flag in the goniometer, but uses the settings
        otherwise) 
        """
        correction = self.goniometer.get_ADS_to_fixed_correction(self.__get_range_theta())
        self.experimental_pattern.apply_correction(correction)

    def convert_to_ads(self):
        """
        Converts the experimental data from fixed slits to ADS in-place 
        (disregards the `has_ads` flag in the goniometer, but uses the settings
        otherwise) 
        """
        correction = 1.0 / self.goniometer.get_ADS_to_fixed_correction(self.__get_range_theta())
        self.experimental_pattern.apply_correction(correction)

    def __get_range_theta(self):
        if len(self.experimental_pattern) <= 1:
            return self.goniometer.get_default_theta_range()
        else:
            return np.radians(self.experimental_pattern.data_x * 0.5)

    def __repr__(self):
        return "Specimen(name='%s')" % self.name

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

SIGNAL_RUN_LAST = None
TYPE_NONE = None
TYPE_PYOBJECT = None

def type_register(self, *args, **kwargs):
    pass


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from mvc import Observer
from pyxrd.generic.models import (
    PyXRDLine,
    CalculatedLine,
    ExperimentalLine,
)

__all__ = [
    'TestPyXRDLine',
    'TestCalculatedLine',
    'TestExperimentalLine',
]

class LineObserver(Observer):
    needs_update_recieved = False
    @Observer.observe("data_changed", signal=True)
    @Observer.observe("visuals_changed", signal=True)
    def on_update_needed(self, model, prop_name, info):
        self.needs_update_recieved = True

class BaseTestLines():

    class BaseTestLine(unittest.TestCase):

        def setUp(self):
            self.line = self.line_type()
            self.observer = LineObserver(model=self.line)

        def tearDown(self):
            self.observer.relieve_model(self.line)
            del self.observer
            del self.line

        def test_not_none(self):
            self.assertIsNotNone(self.line)

        test_lw = create_object_attribute_test('line', 'lw', 5)
        test_color = create_object_attribute_test('line', 'color', '#FF0000')
        test_label = create_object_attribute_test('line', 'label', '#FF0000')

        def _set_some_data(self):
            x = [1, 2, 3, 4, 5, 6, 7 , 8 , 9 , 10, 11, 12 , 13, 14, 15, 16, 17, 18, 19, 20]
            y = [0, 0, 0, 0, 0, 0, 10, 20, 30, 40, 80, 160, 80, 40, 30, 20, 10, 0 , 0 , 0 ]
            y = list(zip(y, y))
            self.line.set_data(x, y)

        def test_updates(self):
            self.observer.needs_update_recieved = False
            self._set_some_data()
            self.assertTrue(self.observer.needs_update_recieved)

        def test_data(self):
            self._set_some_data()
            self.assertEqual(self.line.num_columns, 3)
            self.assertEqual(self.line.max_display_y, 160)
            self.assertEqual(self.line.size, 20)
            self.assertEqual(self.line.get_y_at_x(7), 10)
            self.assertEqual(self.line.get_y_at_x(10.5), 60)

        def test_names(self):
            self.observer.needs_update_recieved = False
            self._set_some_data()
            names = ["TestName"]
            self.line.y_names = names
            self.assertEqual(self.line.get_y_name(0), names[0])

        def test_append_valid(self):
            self.line.append(0, 0)
            self.assertEqual(self.line[0], (0.0, [0.0]))

        def test_append_valid_multi(self):
            self.line.append(0, [0, 1, 2])
            self.assertEqual(self.line[0], (0.0, [0.0, 1.0, 2.0]))

        def test_signal(self):
            self.observer.needs_update_recieved = False
            self.line.lw = 10
            self.assertTrue(self.observer.needs_update_recieved)

        def test_serialisation(self):
            x = [1, 2, 3, 4, 5, 6, 7 , 8 , 9 , 10, 11, 12 , 13, 14, 15, 16, 17, 18, 19, 20]
            y = [0, 0, 0, 0, 0, 0, 10, 20, 30, 40, 80, 160, 80, 40, 30, 20, 10, 0 , 0 , 0 ]
            self.line.set_data(x, y)
            serialised1 = self.line._serialize_data()
            self.line._set_from_serial_data(serialised1)
            serialised2 = self.line._serialize_data()
            self.assertEqual(serialised1, serialised2)

        pass # end of class
        
    pass # end of class

class TestPyXRDLine(BaseTestLines.BaseTestLine):

    line_type = PyXRDLine

    pass # end of class

class TestCalculatedLine(BaseTestLines.BaseTestLine):

    line_type = CalculatedLine

    pass # end of class

class TestExperimentalLine(BaseTestLines.BaseTestLine):

    line_type = ExperimentalLine

    # TODO:
    # test bg substr
    # test smooth
    # test strip
    # test capping -> max intensity

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from ..abstract_adapter import AbstractAdapter

def wrap_property_to_treemodel_type(model, prop, treemodel_type):
    prop_value = getattr(model, prop.label)
    if not isinstance(prop_value, Gtk.TreeModel):
        wrapper = getattr(model, "__%s_treemodel_wrapper" % prop.label, None)
        if wrapper is None or not wrapper.is_wrapping(model, prop.label):
            wrapper = treemodel_type(model, prop)
        setattr(model, "__%s_treemodel_wrapper" % prop.label, wrapper)
        prop_value = wrapper
    return prop_value

def wrap_treenode_property_to_treemodel(model, prop):
    """
        Convenience function that (sparsely) wraps a TreeNode property
        to an ObjectTreeStore. If the property is a Gtk.TreeModel instance,
        it returns it without wrapping.
    """
    from .treemodels import ObjectTreeStore
    return wrap_property_to_treemodel_type(model, prop, ObjectTreeStore)

def wrap_list_property_to_treemodel(model, prop):
    """
        Convenience function that (sparsely) wraps a list property
        to an ObjectListStore. If the property is an Gtk.TreeModel instance,
        it returns it without wrapping.
    """
    from .treemodels import ObjectListStore
    return wrap_property_to_treemodel_type(model, prop, ObjectListStore)

def wrap_xydata_to_treemodel(model, prop):
    """
        Convenience function that (sparsely) wraps an XYData model
        to an XYListStore. If the property is an Gtk.TreeModel instance,
        it returns it without wrapping.
    """
    from .treemodels import XYListStore
    return wrap_property_to_treemodel_type(model, prop, XYListStore)

class AbstractTreeViewAdapter(AbstractAdapter):
    """
        Abstract base class for the ObjectTreeViewAdapter and
        XYTreeViewAdapter.
    """
    toolkit = "gtk"
    _check_widget_type = Gtk.TreeView

    _signal = "changed"

    def __init__(self, controller, prop, widget):
        super(AbstractTreeViewAdapter, self).__init__(controller, prop, widget)
        if self._check_widget_type is not None:
            widget_type = type(widget)
            if not isinstance(widget, self._check_widget_type):
                raise TypeError("A '%s' can only be used for (a subclass of) a '%s' widget, and not for a '%s'!" % (
                       type(self), self._check_widget_type, widget_type
                ))
        self._connect_widget()

    def _connect_widget(self):
        self._widget.set_model(self._treestore)
        setup = getattr(self._controller, "setup_%s_tree_view" % self._prop.label, None)
        if callable(setup):
            setup(self._treestore, self._widget)
        else:
            logger.error("Could not find setup callable for tree view widget '%s' adapted to '%s'" % (
                self._widget,
                self._prop.label
            ))

    def _disconnect_widget(self, widget=None):
        # TODO reset_tree_view support
        self._widget.set_model(None)

    def _connect_model(self):
        pass # nothing to do

    def _disconnect_model(self, model=None):
        pass # nothing to do

    def _read_widget(self):
        pass # nothing to do

    def _write_widget(self, val):
        pass # nothing to do

    def _read_property(self, *args):
        pass # nothing to do

    def _write_property(self, value, *args):
        pass # nothing to do

    pass # end of class

class ObjectListViewAdapter(AbstractTreeViewAdapter):
    """
        An adapter for a TreeView widget, representing a list of objects.
    """

    widget_types = ["object_list_view", ]

    def __init__(self, controller, prop, widget):
        assert hasattr(prop, "data_type"), "ObjectTreeViewAdapter requires the " + \
            "'data_type' attribute to be set on the property descriptor.\n" + \
            "Controller: '%s', Model: '%s', Property: '%s'" % (controller, controller.model, prop.label)
        self._treestore = wrap_list_property_to_treemodel(controller.model, prop)
        super(ObjectListViewAdapter, self).__init__(controller, prop, widget)

    pass # end of class

class XYListViewAdapter(AbstractTreeViewAdapter): # TODO move this back outside this namespace or move the XYData object back in here...
    """
        An adapter for a TreeView widget, representing an XYData model.
    """

    widget_types = ["xy_list_view", ]

    def __init__(self, controller, prop, widget):
        self._treestore = wrap_xydata_to_treemodel(controller.model, prop)
        super(XYListViewAdapter, self).__init__(controller, prop, widget)

    pass # end of class

class ObjectTreeViewAdapter(AbstractTreeViewAdapter):
    """
        An adapter for a TreeView widget, representing a tree of objects.
    """

    widget_types = ["object_tree_view", ]

    def __init__(self, controller, prop, widget):
        self._treestore = wrap_treenode_property_to_treemodel(controller.model, prop)
        super(ObjectTreeViewAdapter, self).__init__(controller, prop, widget)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Tobias Weber
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .base import ObsWrapperBase

class ObsWrapper (ObsWrapperBase):
    """
    Base class for wrappers, like user-classes and sequences.
    Use sub-classes!
    """

    def __init__(self, obj, method_names):
        ObsWrapperBase.__init__(self)

        self._obj = obj
        self.__doc__ = obj.__doc__

        # Creates a derived class which is a singleton which self is
        # going to be an instance of. All method_names are then
        # wrapped within it.
        # See http://stackoverflow.com/questions/1022499/emulating-membership-test-in-python-delegating-contains-to-contained-object
        d = dict((name, self.__get_wrapper(name)) for name in method_names)
        self.__class__ = type(self.__class__.__name__, (self.__class__,), d)
        return

    def __get_wrapper(self, name):
        def _wrapper_fun(self, *args, **kwargs):
            self._notify_method_before(self._obj, name, args, kwargs)
            res = getattr(self._obj, name)(*args, **kwargs)
            self._notify_method_after(self._obj, name, res, args, kwargs)
            return res
        return _wrapper_fun

    # For all fall backs
    def __getattr__(self, name): return getattr(self._obj, name)
    def __repr__(self): return self._obj.__repr__()
    def __str__(self): return self._obj.__str__()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from collections import namedtuple

from ....observers import Observer
from ....models.xydata import XYData

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GLib, GObject  # @UnresolvedImport

from .base_models import BaseObjectListStore

class PointMeta():
    @classmethod
    def get_column_properties(cls):
        return [
            ('x', float),
            ('y', float)
        ]
Point = namedtuple('Point', ['x', 'y'])
Point.Meta = PointMeta

class XYListStore(BaseObjectListStore, Observer):
    """
        GenericTreeModel implementation that wraps an XYData model.
    """
    _model = None
    _prop_name = None
    _last_lenght = 0

    __gsignals__ = {
        'columns-changed' : (GObject.SignalFlags.RUN_LAST, None, ())  # @UndefinedVariable
    }

    @property
    def _data(self):
        return getattr(self._model, self._prop_name, None)

    def is_wrapping(self, model, prop_name):
        return self._model == model and self._prop_name == prop_name

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, model, prop):
        # Continue initialisation:
        BaseObjectListStore.__init__(self, Point)
        
        # Check this really is an XYData property:
        self._flush()
        self._model = model
        self._prop_name = prop.label
        _data = getattr(self._model, self._prop_name, None)
        assert isinstance(_data, XYData), \
            "Can only wrap XYData (or subclasses) instances to a XYListStore," + \
            "but got '%s' instead from property '%s' on model '%s'." % (
                _data, self._prop_name, self._model)
        
        Observer.__init__(self, model=self._data)
        self.set_property("leak-references", False)

        self._last_length = len(self)
        self._last_num_col = self._data.num_columns

        # Force update:
        self._emit_update()

    @Observer.observe("data_changed", signal=True)
    def on_data_changed(self, model, name, info):
        if model == self._data:
            self._emit_update()

    def _emit_update(self):
        # Invalidate iters, we're (probably) changing stuff:
        self._schedule_flush()

        # 1. check if number of columns has changed since last update
        #    if it has changed, emit the corresponding event
        if self._last_num_col != self._data.num_columns:
            self.emit("columns-changed")
            self._last_num_col = self._data.num_columns

        # 2. check if length has changed, if shorter emit removed signals
        #    for the lost elements, if longer emit insert signals
        row_diff = len(self._data) - self._last_length
        if row_diff > 0:
            for i in range(self._last_length, self._last_length + row_diff, 1):
                path = self.on_get_path(i)
                itr = self.get_iter(path)
                self.row_inserted(path, itr)
        elif row_diff < 0:
            for i in range(self._last_length, self._last_length + row_diff - 1, -1):
                path = self.on_get_path(i)
                self.row_deleted(path)
        self._last_length = len(self._data)

        # 3. Emit row-changed signals for all other rows:
        for i in range(0, len(self._data)):
            path = self.on_get_path(i)
            itr = self.get_iter(path)
            self.row_changed(path, itr)

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def on_get_flags(self):
        return Gtk.TreeModelFlags.LIST_ONLY

    def on_get_iter(self, path): # returns a rowref, they're actually just paths
        if hasattr(path, "get_indices"):
            path = path.get_indices()
        sp = ":".join(map(lambda i: "%d" % i, path))
        if not sp in self._cache:
            try:
                i = path[0]
                if i >= 0 and i < len(self):
                    self._cache[sp] = [i, ]
            except IndexError:
                pass
        return self._cache.get(sp, None)
        self._schedule_flush()
        return

    def _schedule_flush(self):
        if not self._flush_scheduled:
            def idle_add():
                GLib.idle_add(self._flush)
                return False # delete timeout

            GLib.timeout_add(500, idle_add)
            self._flush_scheduled = True

    def _flush(self):
        self.invalidate_iters()
        self._cache = {} # del _cache - keep no ref to this dict
        self._flush_scheduled = False
        return False # In case we are called from idle signal

    def on_get_value(self, rowref, column):
        if column == self.c_x:
            return self._data.data_x[rowref[0]]
        elif column >= self.c_y:
            return self._data.data_y[rowref[0], column - 1]
        else:
            raise AttributeError

    def on_get_path(self, rowref): # rowrefs are paths, unless they're None
        if rowref is None:
            return None
        if isinstance(rowref, tuple):
            return rowref
        elif isinstance(rowref, list):
            return tuple(rowref)
        else:
            return rowref,

    def on_iter_next(self, rowref):
        if rowref is not None:
            itr = self.on_get_iter((rowref[0] + 1,))
            return itr
        else:
            return None

    def on_iter_children(self, rowref):
        if rowref is not None:
            return None
        elif len(self) > 0:
            return self.on_get_iter((0,))
        return None

    def on_iter_has_child(self, rowref):
        if rowref is not None:
            return False
        elif len(self) > 0:
            return True
        return False

    def on_iter_n_children(self, rowref):
        if rowref is not None:
            return 0
        return len(self)

    def on_iter_nth_child(self, rowref, n):
        if rowref is not None:
            return None
        if n < 0 or n >= len(self):
            return None
        return self.on_get_iter((n,))

    def on_iter_parent(self, rowref):
        return None

    def on_get_n_columns(self):
        return self._data.num_columns

    def on_get_column_type(self, index):
        return float

    def __len__(self):
        return self._data.size

    pass # end of class

GObject.type_register(XYListStore)  # @UndefinedVariable



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
    This file contains some flags which (partially) control if certain tests
    are run. This way lengthy tests can be skipped in the default set.
"""

SKIP_REFINEMENT_TEST = True



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, unittest

from pyxrd.file_parsers.base_parser import BaseParser

__all__ = [
    'TestParserMixin',
]


def load_data_from_files(*files):
    basepath = os.path.realpath(os.getcwd())
    for fname in files:
        with open(basepath + "/" + fname, 'rb') as fp:
            yield fp

class BaseTestParsers(object):

    class BaseTestParser(unittest.TestCase):

        parser_class = BaseParser
        file_data = [
            "",
        ]

        def test_description(self):
            self.assertNotEqual(self.parser_class.description, "")

        def test_filters(self):
            self.assertIsNotNone(self.parser_class.file_filter)

        def test_parsing(self):
            for fp in self.file_data:
                data_objects = self.parser_class.parse(fp)
                self.assertGreater(len(data_objects), 0)

        # TODO:
        # - check arguments such as close.



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

# FIXME clean this mess up

"""
Shortcuts are provided to the following classes defined in submodules:

.. class:: Model
   :noindex:
.. class:: TreeStoreModel
   :noindex:
.. class:: ListStoreModel
   :noindex:
.. class:: TextBufferModel
   :noindex:
.. class:: ModelMT
   :noindex:
.. class:: Controller
   :noindex:
.. class:: View
   :noindex:
.. class:: Observer
   :noindex:
.. class:: Observable
   :noindex:

"""

from .support.version import LooseVersion 

from .__version import __version__

# Class shortcuts:
from .observers import Observer
from .support.observables import Signal, Observable
from .models import *

try:
    from .view import View
    from .controller import Controller
except ImportError:
    import logging
    logging.getLogger(__name__).warning("ImportError when trying to load View and/or Controller: do you have PyGTK/GObject installed?")



# (c) 2003 Gustavo J A M Carneiro gjc at inescporto.pt
#     2004-2005 Filip Van Raemdonck
#
# http://www.daa.com.au/pipermail/pygtk/2003-August/005775.html
# Message-ID: <1062087716.1196.5.camel@emperor.homelinux.net>
#     "The license is whatever you want."
#
# This file was downloaded from http://www.sysfs.be/downloads/
# Adaptions 2009-2010 by Martin Renold:
# - let KeyboardInterrupt through
# - print traceback to stderr before showing the dialog
# - nonzero exit code when hitting the "quit" button
# - suppress more dialogs while one is already active
# - fix Details button when a context in the traceback is None
# - remove email features
# - fix lockup with dialog.run(), return to mainloop instead
# see also http://faq.pyGtk.org/index.py?req=show&file=faq20.010.htp
# Changes 2012 by Mathijs Dumon:
# - removed the gtkcompat import statement

import inspect, linecache, pydoc, sys, traceback
from io import StringIO
from gettext import gettext as _

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, Gdk, Pango  # @UnresolvedImport

original_excepthook = sys.excepthook

class GtkExceptionHook():
    """
        Exception handling GTK-dialog hook 
    """

    # Function that will be called when the user presses "Quit"
    # Return True to confirm quit, False to cancel
    quit_confirmation_func = None

    exception_dialog_active = False

    RESPONSE_QUIT = 1

    def analyze_simple (self, exctyp, value, tb):
        """
            Analyzes the exception into a human readable stack trace
        """
        trace = StringIO()
        traceback.print_exception (exctyp, value, tb, None, trace)
        return trace

    def lookup (self, name, frame, lcls):
        '''Find the value for a given name in the given frame'''
        if name in lcls:
            return 'local', lcls[name]
        elif name in frame.f_globals:
            return 'global', frame.f_globals[name]
        elif '__builtins__' in frame.f_globals:
            builtins = frame.f_globals['__builtins__']
            if type (builtins) is dict:
                if name in builtins:
                    return 'builtin', builtins[name]
            else:
                if hasattr (builtins, name):
                    return 'builtin', getattr (builtins, name)
        return None, []

    _parent_window = [None]
    _level = 0
    @property
    def parent_window(self):
        return self._parent_window[self._level]
    @parent_window.setter
    def parent_window(self, value):
        self._parent_window[self._level] = value

    def overriden_parent_window(self, parent):
        """
            Sets the parent window temporarily to another value
            and returns itself, so this can be used as a context
            manager.
        """
        self._parent_window.append(parent)
        self._level = len(self._parent_window)
        return self

    def __enter__(self):
        pass

    def __exit__(self):
        if self._level > 0:
            self._parent_window.pop(self._level)
            self._level = len(self._parent_window)

    def analyze (self, exctyp, value, tb):
        """
            Analyzes the exception into a human readable stack trace
        """
        import tokenize, keyword

        trace = StringIO()
        nlines = 3
        frecs = inspect.getinnerframes (tb, nlines)
        trace.write ('Traceback (most recent call last):\n')
        for frame, fname, lineno, funcname, context, _ in frecs:
            trace.write ('  File "%s", line %d, ' % (fname, lineno))
            args, varargs, varkw, lcls = inspect.getargvalues (frame)

            def readline (lno=[lineno], *args):
                if args: print(args)
                try: return linecache.getline (fname, lno[0])
                finally: lno[0] += 1
            all, prev, name, scope = {}, None, '', None
            for ttype, tstr, stup, etup, line in tokenize.generate_tokens (readline):
                if ttype == tokenize.NAME and tstr not in keyword.kwlist:
                    if name:
                        if name[-1] == '.':
                            try:
                                val = getattr (prev, tstr)
                            except AttributeError:
                                # XXX skip the rest of this identifier only
                                break
                            name += tstr
                    else:
                        assert not name and not scope
                        scope, val = self.lookup(tstr, frame, lcls)
                        name = tstr
                    if val is not None:
                        prev = val
                elif tstr == '.':
                    if prev:
                        name += '.'
                else:
                    if name:
                        all[name] = (scope, prev)
                    prev, name, scope = None, '', None
                    if ttype == tokenize.NEWLINE:
                        break

            try:
                details = inspect.formatargvalues (args, varargs, varkw, lcls, formatvalue=lambda v: '=' + pydoc.text.repr (v))
            except:
                # seen that one on Windows (actual exception was KeyError: self)
                details = '(no details)'
            trace.write (funcname + details + '\n')
            if context is None:
                context = ['<source context missing>\n']
            trace.write (''.join (['    ' + x.replace ('\t', '  ') for x in [a for a in context if a.strip()]]))
            if len (all):
                trace.write ('  variables: %s\n' % str (all))

        trace.write ('%s: %s' % (exctyp.__name__, value))
        return trace

    def __call__ (self, exctyp, value, tb):
        """
            This is called when an exception occurs.
        """
        if exctyp is KeyboardInterrupt:
            return original_excepthook(exctyp, value, tb)
        sys.stderr.write(self.analyze_simple (exctyp, value, tb).getvalue())
        if self.exception_dialog_active:
            return

        Gdk.pointer_ungrab(Gdk.CURRENT_TIME)  # @UndefinedVariable
        Gdk.keyboard_ungrab(Gdk.CURRENT_TIME)  # @UndefinedVariable

        self.exception_dialog_active = True
        # Create the dialog
        dialog = Gtk.MessageDialog(
            parent=self.parent_window,
            flags=0, type=Gtk.MessageType.WARNING, buttons=Gtk.ButtonsType.NONE
        )
        dialog.set_title (_("Bug Detected"))

        primary = _("<big><b>A programming error has been detected.</b></big>")
        secondary = _("It probably isn't fatal, but the details should be reported to the developers nonetheless.")

        try:
            setsec = dialog.format_secondary_text
        except AttributeError:
            raise
            dialog.vbox.get_children()[0].get_children()[1].set_markup ('%s\n\n%s' % (primary, secondary))
        else:
            del setsec
            dialog.set_markup (primary)
            dialog.format_secondary_text (secondary)

        dialog.add_button (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        dialog.add_button (Gtk.STOCK_QUIT, self.RESPONSE_QUIT)

        # Add an expander with details of the problem to the dialog
        def expander_cb(expander, *ignore):
            # Ensures that on deactivating the expander, the dialog is resized down
            if expander.get_expanded():
                dialog.set_resizable(True)
            else:
                dialog.set_resizable(False)
        details_expander = Gtk.Expander()
        details_expander.set_label(_("Details..."))
        details_expander.connect("notify::expanded", expander_cb)

        textview = Gtk.TextView(); textview.show()
        textview.set_editable (False)
        textview.modify_font (Pango.FontDescription ("Monospace"))

        sw = Gtk.ScrolledWindow(); sw.show()
        sw.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.set_size_request(800, 400)
        sw.add (textview)

        details_expander.add (sw)
        details_expander.show_all()
        dialog.get_content_area().pack_start(details_expander, True, True, 0)

        # Get the traceback and set contents of the details
        try:
            trace = self.analyze(exctyp, value, tb).getvalue()
        except:
            trace = _("Exception while analyzing the exception.")
        buf = textview.get_buffer()
        buf.set_text (trace)

        # Connect callback and present the dialog
        dialog.connect('response', self._dialog_response_cb, trace)
        dialog.set_modal(True)
        dialog.show()
        dialog.present()


    def _dialog_response_cb(self, dialog, resp, trace):

        if resp == self.RESPONSE_QUIT and Gtk.main_level() > 0:
            if not callable(self.quit_confirmation_func):
                sys.exit(1) # Exit code is important for IDEs
            else:
                if self.quit_confirmation_func():
                    sys.exit(1) # Exit code is important for IDEs
                else:
                    dialog.destroy()
                    self.exception_dialog_active = False

        else:
            dialog.destroy()
            self.exception_dialog_active = False

def plugin_gtk_exception_hook():
    hook = GtkExceptionHook()
    sys.excepthook = hook
    return hook



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import zipfile

from pyxrd.generic.io import get_case_insensitive_glob, COMPRESSION

from pyxrd.file_parsers.json_parser import JSONParser
from pyxrd.file_parsers.registry import ParserNamespace

goniometer_parsers = ParserNamespace("GoniometerParsers")

@goniometer_parsers.register_parser()
class JSONGoniometerParser(JSONParser):

    description = "Goniometer file *.GON"
    extensions = get_case_insensitive_glob("*.GON")
    mimetypes = ["application/octet-stream", "application/zip"]

    pass #end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

__all__ = [
    'AbstractTestProbModel',
]

class AbstractTestProbModel():

    prob_model = None
    prob_model_type = None

    def setUp(self):
        self.prob_model = self.prob_model_type()

    def tearDown(self):
        del self.prob_model

    def test_not_none(self):
        self.assertIsNotNone(self.prob_model)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from math import pi
import logging
logger = logging.getLogger(__name__)

def get_atomic_scattering_factor(angstrom_range, atom_type):
    r"""
    Calculates the atomic scatter factor for 
        - angstrom_range: numpy array of (2*sin(θ) / λ)² values (= 1/A² units)
        - atom_type: an :class:`~pyxrd.calculations.data_objects.AtomTypeData`
          instance
          
    The atomic scattering factor is calculated as follows:
    
    .. math::
        :nowrap:
        
        \begin{flalign*}
            & ASF = \left[ c + \sum_{i=1}^{5}{ \left( a_i \cdot e ^ { - b_i \cdot {\left(2 \cdot \frac{sin(\theta)}{\lambda}\right)}^2 } \right) } \right] 
            \cdot e ^ { B \cdot {\left(2 \cdot \frac{sin(\theta)}{\lambda}\right)}^2 }  
        \end{flalign*}
        
    Where a_i, b_i and c are the scattering factors taken from `atom_type`
    """
    if atom_type is not None:
        asf = np.sum(atom_type.par_a * np.exp(-atom_type.par_b * angstrom_range[..., np.newaxis]), axis=1) + atom_type.par_c
        asf = asf * np.exp(-atom_type.debye * angstrom_range)
        return asf
    else:
        logger.warning("get_atomic_scattering_factor reports: 'None found!'")
        return np.zeros_like(angstrom_range)

def get_structure_factor(range_stl, atom):
    r"""
    Calculates the atom's structure factor for
        - range_stl: numpy array of 2*sin(θ) / λ values (= 1/nm units)
        - atom_type: an :class:`~pyxrd.calculations.data_objects.AtomData`
          instance 
          
    The structure factor is calculated using the atomic scattering factor ASF
    (calculated by :meth:`~pyxrd.calculations.atoms.get_atomic_scattering_factor`)
    as follows:

    .. math::
        :nowrap:
        
        \begin{flalign*}
            & SF = ASF \cdot p \cdot e ^ { 2 \cdot \pi \cdot z \cdot i \cdot \frac{2 \cdot sin(\theta)}{\lambda} }
        \end{flalign*}
       
    """
    if atom is not None and atom.atom_type is not None:
        angstrom_range = ((range_stl * 0.05) ** 2)

        asf = get_atomic_scattering_factor(angstrom_range, atom.atom_type)

        return asf * atom.pn * np.exp(2 * pi * atom.z * range_stl * 1j)
    else:
        logger.warning("get_structure_factor reports: 'None found!'")
        return np.zeros_like(range_stl)



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import weakref
import types

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from .generic_tree_model import GenericTreeModel

class BaseObjectListStore(GenericTreeModel):
    """
        Base mixin for creating GenericTreeModel implementations for lists of
        objects. It maps the columns of the store with properties of the object.
        If the PyGTK modules are not available (e.g. on a headless HPC cluster),
        a dummy gtk module is loaded. Not all (GTK) functionality is enabled in that case.
    """

    # PROPERTIES
    _columns = None # list of tuples (name, type)
    _class_type = None

    __weakref_model = None
    @property
    def _model(self):
        if self.__weakref_model is not None:
            return self.__weakref_model()
        else:
            return None
    @_model.setter
    def _model(self, value):
        if value is not None:
            self.__weakref_model = weakref.ref(value)
        else:
            self.__weakref_model = None


    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, class_type):
        GenericTreeModel.__init__(self)
        self.set_property("leak-references", False)
        if class_type is None:
            raise ValueError(
                 'Invalid class_type for %s! Expecting object, but None was given' 
                 % type(self)
             )
        elif not hasattr(class_type, "Meta"):
            raise ValueError(
                 'Invalid class_type for `%s`! `%s` does not have a `Meta` attribute!' 
                 % (type(self), class_type)
             )
        elif not hasattr(class_type.Meta, 'get_column_properties'):
            raise ValueError(
                 'Invalid class_type for %s! %s.Meta does not have get_column_properties method!' 
                 % (type(self), class_type)
             )
        else:
            self.setup_class_type(class_type)

    def setup_class_type(self, class_type):
        self._class_type = class_type
        self._columns = []
        for item in self._class_type.Meta.get_column_properties():
            title, col_type = item
            if isinstance(col_type, str):
                col_type = 'gchararray'
            # TODO map other types we might encounter...
            self._columns.append((title, col_type))

        i = 0
        for col in self._columns:
            setattr(self, "c_%s" % col[0], i)
            i += 1

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def on_get_flags(self):
        return Gtk.TreeModelFlags.LIST_ONLY | Gtk.TreeModelFlags.ITERS_PERSIST

    def on_get_n_columns(self):
        return len(self._columns)

    def on_get_column_type(self, index):
        return self._columns[index][1]

    def get_user_data_from_path(self, path):
        return self.on_get_iter(path)

    def convert(self, col, new_val):
        if isinstance(self._columns[col][1], str):
            return str(new_val)
        else:
            return self._columns[col][1](new_val)

    def get_objects(self):
        raise NotImplementedError()

    def iter_objects(self):
        raise NotImplementedError()

    def __reduce__(self):
        raise NotImplementedError()

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest
from mock import Mock

from test.utils import create_object_attribute_test

from pyxrd.atoms.models import Atom

__all__ = [
    'TestAtom',
]

class TestAtom(unittest.TestCase):

    atom_type = None

    def setUp(self):
        self.atom = Atom()

    def tearDown(self):
        del self.atom

    def get_mocked_hierarchy(self):       
        oxygen = Mock()
        oxygen.name = "O1-"
        hydrogen = Mock()
        hydrogen.name = "H+"
                
        project = Mock()
        project.atom_types = [oxygen, hydrogen]
        phase = Mock()
        phase.attach_mock(project, 'project')
        component = Mock()
        component.attach_mock(phase, 'phase')
        
        return oxygen, hydrogen, component, phase, project

    def test_not_none(self):
        self.assertIsNotNone(self.atom)

    def test_data_object(self):
        self.assertIsNotNone(self.atom.data_object)

    test_name = create_object_attribute_test("atom", "name", "Test Name")
    test_pn = create_object_attribute_test("atom", "pn", 3)
    test_default_z = create_object_attribute_test("atom", "default_z", 5.3)
    test_stretch_values = create_object_attribute_test("atom", "stretch_values", True)

    def test_parent(self):
        parent_atom = Atom(name="Parent")
        self.atom.parent = parent_atom
        self.assertEqual(self.atom.parent, parent_atom)

    def test_set_atom_type(self):
        oxygen, hydrogen, component, _, _ = self.get_mocked_hierarchy()
        atom = Atom(
            parent = component,
            name = "O",
            atom_type = oxygen,
        )
        atom.atom_type = hydrogen
        self.assertEqual(atom.atom_type, hydrogen)

    def test_z_calculations(self):
        # Checks wether the atom can calculate stretched values:
        # 1. When everything is set up the way it should be:
        default_z = 9.0
        lattice_d = 5.4
        factor = 0.5
        
        parent = Mock()
        parent.configure_mock(**{
            'get_interlayer_stretch_factors.return_value': (lattice_d, factor)
        })
        
        atom = Atom(parent=parent)
        atom.stretch_values = True
        atom.default_z = default_z
        z = atom.z
        self.assertEqual(z, lattice_d + (default_z - lattice_d) * factor)
        # 2. When no component is set, but stretched is True: should not raise an error, but simple ignore the stretching
        atom.parent = None
        z = atom.z

    def test_structure_factors(self):
        import numpy as np
        rng = 2.0 * np.sin(np.arange(30)) / 0.154056
        res = self.atom.get_structure_factors(rng)
        self.assertIsNotNone(res)
        
    def test_loads_atom_type_by_name(self):
        atom_json_dict = {
            "uuid": "878341b04e9e11e2b238150ae229a525", 
            "name": "O", 
            "default_z": 0.66, 
            "pn": 6.0, 
            "atom_type_name": "O1-"
        }
        
        oxygen, _, component, _, _ = self.get_mocked_hierarchy()
        
        atom = Atom.from_json(parent = component, **atom_json_dict)
        atom.resolve_json_references()
        self.assertEqual(atom.atom_type, oxygen)
        

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .labeled_property import LabeledProperty
from .signal_property import SignalProperty
from .cast_property import CastProperty
from .cast_choice_property import CastChoiceProperty
from .list_property import ListProperty

from .uuid_property import UUIDProperty

from .bool_property import BoolProperty
from .integer_properties import IntegerChoiceProperty, IntegerProperty
from .float_properties import FloatChoiceProperty, FloatProperty
from .string_properties import StringChoiceProperty, StringProperty, ColorProperty

from .signal_mixin import SignalMixin
from .read_only_mixin import ReadOnlyMixin
from .action_mixins import GetActionMixin, SetActionMixin
from .observe_mixin import ObserveMixin


__all__ = [
    "LabeledProperty",
    "SignalProperty",
    "CastProperty",
    "CastChoiceProperty",
    "ListProperty",
    "UUIDProperty",
    "BoolProperty",
    "IntegerChoiceProperty", "IntegerProperty",
    "FloatChoiceProperty", "FloatProperty",
    "StringChoiceProperty", "StringProperty", "ColorProperty",
    "SignalMixin",
    "ReadOnlyMixin",
    "GetActionMixin", "SetActionMixin",
    "ObserveMixin",
]


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from pkg_resources import resource_filename # @UnresolvedImport
from warnings import warn

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

from mvc.view import View
from mvc.adapters.gtk_support.widgets import ScaleEntry

from pyxrd.data import settings
from pyxrd.generic.mathtext_support import get_string_safe

class BaseView(View):
    """
        Basic view providing some common code
        TODO attribute docs!
    """
    builder = ""
    modal = False
    resizable = True
    widget_format = "%s"
    container_format = "container_%s"

    # A mapping of layout states and actions to be taken when switching to them.
    # Actions should be of the form `action::widget_group`, where the action is one
    # of either hide or show and the widget_group is a key in the `widget_groups`
    # mapping, with the value being the list of widgets to apply the action on to.
    current_layout_state = "FULL"
    layout_state_actions = {
        'FULL': [
            'show::full_mode_only'
        ],
        'VIEWER': [
            'hide::full_mode_only'
        ]
    }
    widget_groups = { }

    def __init__(self, *args, **kwargs):
        super(BaseView, self).__init__(*args, **kwargs)
        self.parent = kwargs.get("parent", None)
        top = self.get_toplevel()
        if isinstance(top, Gtk.Window):
            top.set_resizable(self.resizable)
            top.set_modal(self.modal)
        if self.parent:
            self.set_layout_mode(self.parent.current_layout_state)

    def create_mathtext_widget(self, text, fallback_text=""):
        # TODO move these to a separate controller namespace module!
        try:
            from pyxrd.generic.mathtext_support import create_image_from_mathtext
            widget = create_image_from_mathtext(text)
        except:
            if fallback_text: text = fallback_text
            widget = Gtk.Label(label=text)
            widget.set_use_markup(True)
            widget.set_justify(Gtk.Justification.CENTER)
        return widget

    def _get_widget_container(self, prop):
        return self[self.container_format % prop.label]

    def _add_widget_to_container(self, widget, container):
        if container is not None:
            child = container.get_child()
            if child is not None:
                container.remove(child)
            container.add(widget)
        widget.show_all()
        return widget

    def add_widget(self, prop, *args, **kwargs):
        method = {
           str: self.add_entry_widget,
           float: self.add_scale_widget,
        }.get(prop.data_type, None)
        if method is not None:
            return method(prop, *args, **kwargs)

    def add_scale_widget(self, prop, enforce_range=True):
        # Create the widget:        
        inp = ScaleEntry(prop.minimum, prop.maximum, enforce_range=enforce_range)
        inp.set_tooltip_text(prop.title)
        self[self.widget_format % prop.label] = inp
            
        # Add & return the widget
        return self._add_widget_to_container(inp, self._get_widget_container(prop))

    def add_entry_widget(self, prop):
        # Create the widget:        
        inp = Gtk.Entry()
        inp.set_tooltip_text(prop.title)
        self[self.widget_format % prop.label] = inp

        # Add & return the widget
        return self._add_widget_to_container(inp, self._get_widget_container(prop))

    def create_input_table(self, table, props, num_columns = 1, widget_callbacks=[]):
        """
            Places widgets (returned by the widget_callbacks) in the given table
            for each property in props. The widget groups (i.e. list of widgets 
            returned by the callbacks for a single property) can be placed in a
            multi-column layout by increasing the num_columns value.
        """
        # total number of properties
        num_props = len(props)
        # number of widgets in each column
        column_width = len(widget_callbacks)
        # 2D array of widgets (each row is a group of widgets returned for a property)
        widgets = [[None for _ in range(column_width)] for _ in range(num_props)]  
        
        # The actual number of rows needed in the table
        num_table_rows = int(num_props / num_columns)
        # The actual number of columns needed in the table
        num_table_columns = num_columns*column_width
        
        # Resize the table widget accordingly
        table.resize(num_table_rows,  num_table_columns)          

        for i, prop in enumerate(props): # i = 0, 1, 2, 3, ...
            # Calculate the column: column_index = 0, cw, 2*cw, 3*cw, ...
            column_index = (i % num_columns) * column_width                
            for widget_index in range(column_width):
                # Create the widget:
                widget = widget_callbacks[widget_index](prop)
                
                # Store it for the return value:
                widgets[i][widget_index] = widget
                
                # Calculate where to place the widget:
                column_offset = column_index + widget_index
                row_offset = int(i / num_columns)
                
                # Attach it to the table               
                table.attach(
                    widget,
                    column_offset, column_offset + 1,
                    row_offset, row_offset + 1,
                    xpadding=2, ypadding=2
                )
        return widgets

    def set_layout_mode(self, state):
        self.current_layout_state = state
        for action in self.layout_state_actions.get(state, []):
            parts = tuple(action.split("::", 1))
            command, group_name = parts
            widgets = []
            if group_name != "all":
                widget_names = self.widget_groups.get(group_name, [])
                widgets = map(lambda name: self[name], widget_names)
            else:
                widgets = self._builder.get_objects()

            if command == "show":
                for widget in widgets:
                    try:
                        widget.set_no_show_all(False)
                        widget.show_all()
                    except AttributeError:
                        pass
            elif command == "hide":
                for widget in widgets:
                    try:
                        widget.set_no_show_all(True)
                        widget.set_visible(False)
                    except AttributeError:
                        pass
            else:
                raise ValueError("Unknown layout state command `%s`!" % command)

    def show_all(self, *args, **kwargs):
        self.show(*args, **kwargs)

    def present(self):
        toplevel = self.get_toplevel()
        toplevel.set_resizable(self.resizable)
        toplevel.set_modal(self.modal)
        toplevel.show_all()
        toplevel.present()
        self.show()

    def get_toplevel(self):        
        for w in [self.top,] + list(self):
            try:
                return self[w].get_toplevel()
            except AttributeError:
                pass
            else:
                break # just for ref

class TitleView(BaseView):
    """
        Mix-in that provides title support for views.
        The class attribute 'title' can be set, if so, this class will
        attempt to set the title attribute upon initialization unless it is 'None'. 
    """
    title = None

    def __init__(self, *args, **kwargs):
        super(TitleView, self).__init__(*args, **kwargs)
        if self.title is not None: self.set_title(self.title)

    def set_title(self, title):
        self.title = title
        self.get_toplevel().set_title(title)

    pass # end of class

class FormattedTitleView(TitleView):
    """
        Mix-in that provides a formatted title support for views.
        The 'title_format' class attribute should be set to a string format
        containing a single string (%s) specifier. When set_title is called, 
        only that part of the string is updated.
    """
    title_format = "%s"
    title = ""

    def set_title(self, title, *args, **kwargs):
        self.title = title
        self.get_toplevel().set_title(self.title_format % self.title)

    pass # end of class

class HasChildView(object):
    """
        Mixin that provides a function to add childviews to containers
    """
    def _add_child_view(self, new_child, container):
        child = container.get_child()
        if child is not None:
            container.remove(child)
        if isinstance(container, Gtk.ScrolledWindow) and not (type(new_child) in (Gtk.TextView, Gtk.TreeView, Gtk.IconView, Gtk.Viewport)):
            container.add_with_viewport(new_child)
        else:
            container.add(new_child)
        new_child.show_all()
        return new_child

    pass # end of class

class DialogView(HasChildView, TitleView):
    """
        Generalised view for editing stuff with an OK button
    """
    builder = resource_filename(__name__, "glade/edit_dialog.glade")
    top = "window_edit_dialog"
    container_widget = "edit_child_box"

    # These should be overriden by the subclass:
    subview_builder = ""
    subview_toplevel = None

    def __init__(self, container_widget=None, subview_builder=None, subview_toplevel=None, *args, **kwargs):
        self.container_widget = container_widget or self.container_widget
        self.subview_builder = subview_builder or self.subview_builder
        self.subview_toplevel = subview_toplevel or self.subview_toplevel
        super(DialogView, self).__init__(*args, **kwargs)
        self._builder.add_from_file(self.subview_builder)
        self._add_child_view(self[self.subview_toplevel], self[self.container_widget])
        return


class ObjectListStoreViewMixin(HasChildView):
    edit_view = None
    edit_view_container = "vwp_edit_object"

    extra_widget_builder = None
    extra_widget_toplevel = ""

    treeview_widget = "edit_objects_treeview"

    @property
    def treeview(self):
        return self[self.treeview_widget]

    @property
    def load_label(self):
        return self["button_load_object"].get_label()

    @load_label.setter
    def load_label(self, value):
        self["button_load_object"].set_label(value)

    @property
    def save_label(self):
        return self["button_save_object"].get_label()

    @save_label.setter
    def save_label(self, value):
        self["button_save_object"].set_label(value)

    @property
    def extra_widget(self):
        return self.extra_widget_box.get_child()

    @extra_widget.setter
    def extra_widget(self, widget):
        child = self.extra_widget_box.get_child()
        if child:
            self.extra_widget_box.remove(child)
        if widget is not None:
            self["extra_box"].add(widget)

    _none_view = None
    @property
    def none_view(self):
        return NoneView()

    def set_selection_state(self, value):
        """
            Sets the state of the view to correspond with the number of currently
            selected objects. Value is either None or a number indicating the number of selected objects.
        """
        self["button_del_object"].set_sensitive(value is not None)
        self["button_save_object"].set_sensitive(value is not None)

    def __init__(self, edit_view_container=None, display_buttons=True, load_label=None, save_label=None, **kwargs):
        self.edit_view_container = edit_view_container or self.edit_view_container
        self.load_label = load_label or self.load_label
        self.save_label = save_label or self.save_label

        if not display_buttons:
            self["vbox_objects"].remove(self["table_data"])

        self.extra_widget_box = self["extra_box"]
        if self.extra_widget_builder is not None:
            self._builder.add_from_file(self.extra_widget_builder)
            self.extra_widget = self._builder.get_object(self.extra_widget_toplevel)
        return

    _on_sr_id = None
    def on_size_requested(self, *args):
        sr = self.child_view.size_request()
        self[self.edit_view_container].set_size_request(sr.height + 20, -1)

    def set_edit_view(self, view):
        if self._on_sr_id is not None and self.child_view is not None:
            self.child_view.disconnect(self._on_sr_id)
        self.edit_view = view
        self.child_view = view.get_top_widget()
        self._add_child_view(self.child_view, self[self.edit_view_container])
        if isinstance(self[self.edit_view_container], Gtk.ScrolledWindow):
            sr = self.child_view.get_size_request()
            self[self.edit_view_container].set_size_request(sr[0], -1)
            self[self.edit_view_container].set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
            self._on_sr_id = self.child_view.connect("size-allocate", self.on_size_requested)


        return self.edit_view

class ObjectListStoreView(ObjectListStoreViewMixin, DialogView):
    """
        Generalised view for editing objects inside an ObjectListStore (using a customisable child view)
         - Standalone version (inside a DialogView)
    """
    subview_builder = resource_filename(__name__, "glade/object_store.glade")
    subview_toplevel = "edit_object_store"

    def __init__(self, edit_view_container=None, display_buttons=True, load_label=None, save_label=None, **kwargs):
        DialogView.__init__(self, **kwargs)
        ObjectListStoreViewMixin.__init__(self, edit_view_container=edit_view_container, display_buttons=display_buttons, load_label=load_label, save_label=save_label, **kwargs)

class ChildObjectListStoreView(ObjectListStoreViewMixin, BaseView):
    """
        Generalised view for editing objects inside an ObjectListStore (using a customisable child view)
         - Child version (to be embedded by a controller)
    """
    edit_view_container = "frame_object_param"

    builder = resource_filename(__name__, "glade/object_store.glade")
    top = "edit_object_store"

    def __init__(self, edit_view_container=None, display_buttons=True, load_label=None, save_label=None, **kwargs):
        BaseView.__init__(self, **kwargs)
        ObjectListStoreViewMixin.__init__(self, edit_view_container=edit_view_container, display_buttons=display_buttons, load_label=load_label, save_label=save_label, **kwargs)

        self["frm_objects_tv"].set_size_request(150, 150)

class InlineObjectListStoreView(BaseView):
    builder = resource_filename(__name__, "glade/inline_ols.glade")
    top = "edit_item"

    @property
    def treeview(self):
        return self['tvw_items']

    @property
    def del_item_widget(self):
        return self['btn_del_item']

    @property
    def add_item_widget(self):
        return self['btn_add_item']

    @property
    def export_items_widget(self):
        return self['btn_export_item']

    @property
    def import_items_widget(self):
        return self['btn_import_item']

    @property
    def type_combobox_widget(self):
        return self['cmb_add_type']

class NoneView(BaseView):
    builder = resource_filename(__name__, "glade/none.glade")
    top = "lbl_caption"
    caption_widget = "lbl_caption"

    def __init__(self, label=None, **kwargs):
        BaseView.__init__(self, **kwargs)
        self._label = self[self.caption_widget]
        if label is not None: self.label = label

    _label = None
    @property
    def label(self):
        return self._label.get_label()

    @label.setter
    def label(self, value):
        self._label.set_label(value)



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from math import sqrt, log

from .math_tools import lognormal

def calculate_distribution(CSDS):
    r"""
    Returns a tuple containing a numpy array with the (log normal) CSDS
    distribution and a float containing the arithmetic mean of that
    distribution.
    
    Takes a single :class:`~CSDSData` object as argument.
    """
    a = CSDS.alpha_scale * log(CSDS.average) + CSDS.alpha_offset
    b = sqrt(CSDS.beta_scale * log(CSDS.average) + CSDS.beta_offset)

    steps = int(CSDS.maximum - CSDS.minimum) + 1

    maxT = 0

    smq = 0
    q_log_distr = []
    TQDistr = dict()
    for i in range(steps):
        T = max(CSDS.minimum + i, 1e-50)
        q = lognormal(T, a, b)
        smq += q

        TQDistr[int(T)] = q
        maxT = T

    TQarr = np.zeros(shape=(maxT + 1,), dtype=float)
    Rmean = 0
    for T, q in TQDistr.items():
        TQarr[T] = q / smq
        Rmean += T * q
    Rmean /= smq

    return TQarr, Rmean



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy.interpolate import interp1d

from .math_tools import mmult

from .CSDS import calculate_distribution
from .goniometer import get_lorentz_polarisation_factor
from .components import get_factors

def get_structure_factors(range_stl, G, comp_list):
    """
        Calculates the structure factor and phase factor for:
        
            - range_stl: numpy array of 2*sin(θ) / λ values
            - G: the number of components (layer types)
            - comp_list: list of 
              :class:`~pyxrd.calculations.data_objects.ComponentData` instances
               
        This function calls :meth:`~pyxrd.calculations.components.get_factors`
        for each component in `comp_list` and stores the returned structure 
        factors and phase difference factors in a numpy array (of type complex)
        with shape (X, G) where X is expanded to fit the shape of `range_stl`.
    """
    shape = range_stl.shape + (G,)
    SF = np.zeros(shape, dtype=np.complex_)
    PF = np.zeros(shape, dtype=np.complex_)
    for i, comp in enumerate(comp_list):
        SF[:, i], PF[:, i] = get_factors(range_stl, comp) # @UndefinedVariable
    return SF, PF

def get_Q_matrices(Q, CSDS_max):
    Qn = np.zeros((CSDS_max + 1,) + Q.shape, dtype=complex)
    Qn[0, ...] = np.copy(Q)
    for n in range(1, CSDS_max + 1):
        Qn[n, ...] = mmult(Qn[n - 1, ...], Q)
    return Qn

def get_absolute_scale(components, CSDS_real_mean, W):
    W = np.diag(W)
    mean_volume = 0.0
    mean_d001 = 0.0
    mean_density = 0.0

    for i, comp in enumerate(components):
        if comp != None:
            mean_volume += comp.volume * W[i]
            mean_d001 += comp.d001 * W[i]
            mean_density += (comp.weight * W[i] / comp.volume)
        else:
            logger.debug("- calc: get_absolute_scale reports: 'Zero observations found!'")

    mean_mass = (CSDS_real_mean * mean_volume ** 2 * mean_density)
    if mean_mass != 0.0:
        return mean_d001 / mean_mass
    else:
        return 0.0

def get_diffracted_intensity(range_theta, range_stl, phase):
    """
        Gets intensity for a single phase without taking the
        lorentz polarization factor into account.
    """
    
    if phase.type == "AbtractPhase":
        raise NotImplementedError
    elif phase.type == "RawPatternPhase":
        return _get_raw_intensity(range_theta, range_stl, phase)
    else:
        return _get_diffracted_intensity(range_theta, range_stl, phase)

def get_intensity(range_theta, range_stl, soller1, soller2, mcr_2theta, phase):
    """
        Gets intensity for a single phase taking the
        lorentz polarization factor into account.
    """

    intensity = get_diffracted_intensity(range_theta, range_stl, phase)
    if phase.apply_lpf:
        return intensity * get_lorentz_polarisation_factor(
            range_theta,
            phase.sigma_star,
            soller1, soller2, mcr_2theta 
        )
    else:
        return intensity

def _get_raw_intensity(range_theta, range_stl, phase):
    assert phase.type == "RawPatternPhase", "Must be RawPatternPhase!"
    f = interp1d(
        phase.raw_pattern_x, phase.raw_pattern_y,
        bounds_error=False, fill_value=0
    )
    i = f(np.rad2deg(2 * range_theta))
    return i

def _get_diffracted_intensity(range_theta, range_stl, phase):
    assert phase.type == "Phase", "Must be Phase!"
    # Check probability model, if invalid return zeros instead of the actual pattern:
    if not phase.valid_probs:
        logger.debug("- calc: get_diffracted_intensity reports: 'Invalid probability found!'")
        return np.zeros_like(range_stl)
    else:
        # Calculate CSDS distribution
        CSDS_arr, CSDS_real_mean = calculate_distribution(phase.CSDS)

        # Get absolute scale
        abs_scale = get_absolute_scale(phase.components, CSDS_real_mean, phase.W)

        # Create a helper function to 'expand' certain arrays, for
        # results which are independent of the 2-theta range
        stl_dim = range_stl.shape[0]
        repeat_to_stl = lambda arr: np.repeat(arr[np.newaxis, ...], stl_dim, axis=0)

        # Repeat junction probabilities & weight fractions
        W = repeat_to_stl(phase.W).astype(np.complex_)
        P = repeat_to_stl(phase.P).astype(np.complex_)

        # Repeat & get SFa and SFb (transpose conjugate) structure factor matrices:
        SF, PF = get_structure_factors(range_stl, phase.G, phase.components)
        SFa = np.repeat(SF[..., np.newaxis, :], SF.shape[1], axis=1)
        SFb = np.transpose(np.conjugate(SFa), axes=(0, 2, 1))

        # Calculate the repetition factor for R+ probabilities:
        rank = P.shape[1]
        reps = rank / phase.G

        # Calculate the structure factor matrix:
        F = np.repeat(np.repeat(np.multiply(SFb, SFa), reps, axis=2), reps, axis=1)

        # Create Q phase factor matrices:
        PF = np.repeat(PF[..., np.newaxis, :], PF.shape[1], axis=1)
        Q = np.multiply(np.repeat(np.repeat(PF, reps, axis=2), reps, axis=1), P)
        Qn = get_Q_matrices(Q, phase.CSDS.maximum)

        # Calculate the intensity:
        sub_total = np.zeros(Q.shape, dtype=np.complex)
        for n in range(phase.CSDS.minimum, phase.CSDS.maximum + 1):
            progression_factor = 0
            for m in range(n + 1, phase.CSDS.maximum + 1):
                progression_factor += (m - n) * CSDS_arr[m]
            sub_total += 2 * progression_factor * Qn[n - 1, ...]

        CSDS_I = repeat_to_stl(np.identity(rank, dtype=np.complex) * CSDS_real_mean)
        sub_total = (CSDS_I + sub_total)
        sub_total = mmult(mmult(F, W), sub_total)
        intensity = np.real(np.trace(sub_total, axis1=2, axis2=1))

        return intensity * abs_scale



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
logger = logging.getLogger(__name__)

from contextlib import contextmanager
from ...adapters.model_adapter import ModelAdapter
from ...support.utils import not_none

class GtkAdapter(ModelAdapter):
    """
        A base class for Gtk-widget Adapters
    """
    toolkit = "gtk"

    # Widget-side value handling:
    _wid_read = None
    _wid_write = None
    _signal = None
    _signal_id = None
    _signal_args = []
    _check_widget_type = None

    def __init__(self, controller, prop, widget,
            prop_read=None, prop_write=None,
            value_error=None, spurious=False,
            wid_read=None, wid_write=None,
            signal=None, signal_args=None, update=True):
        """
            wid_read and wid_write are the methods used for reading and writing the
            widget's value.
            
            signal is the signal name to listen to for widget updates  
            signal_args is the (optional) (list of) argument(s) that will be
            passed when connecting the signal.
    
            Finally, if update is false, the widget will not be initially updated
        """
        super(GtkAdapter, self).__init__(
            controller, prop, widget,
            prop_read=prop_read, prop_write=prop_write,
            value_error=value_error, spurious=spurious
        )
        # Widget-side value handling:
        self._wid_read = not_none(wid_read, self._wid_read)
        self._wid_write = not_none(wid_write, self._wid_write)
        self._signal = not_none(signal, self._signal)
        self._signal_id = None
        self._signal_args = not_none(signal_args, self._signal_args)
        self._update = update

        if self._check_widget_type is not None:
            widget_type = type(widget)
            if not isinstance(widget, self._check_widget_type):
                msg = "Property '%s' from model '%s' has a widget type %s, " \
                      "which can only be used for (a subclass of) a %s " \
                      "widget, and not for a %s widget!" % (
                          prop.label, controller.model, type(self),
                          self._check_widget_type, widget_type
                      )
                raise TypeError(msg)
        # Connect the widget:
        self._connect_widget()

    # ----------------------------------------------------------------------
    #  Widget connecting & disconnecting:
    # ----------------------------------------------------------------------
    def _connect_widget(self):
        """Called when the adapter is ready to connect to the widget"""

        # Connect the widget
        if self._signal:
            try:
                self._signal_id = self._widget.connect(
                    self._signal, self._on_wid_changed, self._signal_args)
            except TypeError:
                logger.error("Failed to connect signal named '%s' on widget '%s' for property '%s'!" % (self._signal, self._widget, self._prop.label))
                raise

        # Updates the widget:
        if self._update: self.update_widget()
        return

    def _on_wid_changed(self, wid, *args):
        """Called when the widget is changed"""
        if self._ignoring_notifs: return
        self.update_model()
        return

    def _disconnect_widget(self, widget=None):
        """Disconnects the widget"""
        if self._signal is not None and self._signal_id is not None:
            widget = not_none(self._widget, widget)
            if widget is not None:
                widget.disconnect(self._signal_id)
            self._signal, self._signal_id = None, None

    # ----------------------------------------------------------------------
    #  Widget-side reading and writing
    # ----------------------------------------------------------------------
    @contextmanager
    def _block_widget_signal(self):
        if self._signal_id is not None: self._widget.handler_block(self._signal_id)
        yield
        if self._signal_id is not None: self._widget.handler_unblock(self._signal_id)

    def _read_widget(self):
        """Returns the value currently stored into the widget."""
        return self._wid_read(self._widget)

    def _write_widget(self, val):
        """Writes value into the widget. If specified, user setter
        is invoked."""
        with self._ignore_notifications():
            return self._wid_write(self._widget, val)

    @staticmethod
    def static_to_class(func):
        def wrapper(c, *args, **kwargs):
            return func(*args, **kwargs)
        return wrapper

    pass # end of class



#!/usr/bin/env python3

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

if __name__ == "__main__":
    # Add parent dir to the path:
    import os, sys
    sys.path.insert(1, os.path.join(sys.path[0], '../..'))    

    from pyxrd.logs import setup_logging
    from pyxrd.server import settings  # @UnusedImport
    setup_logging(basic=True, prefix="PYRO NAMESERVER:")

    import Pyro4.naming
    Pyro4.naming.startNSloop()




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from math import radians
import numpy as np

from mvc.models.properties import (
    LabeledProperty, FloatProperty, IntegerProperty, BoolProperty, StringChoiceProperty,
    SignalMixin, ReadOnlyMixin, ObserveMixin
)

from pyxrd.data import settings

from pyxrd.refinement.refinables.properties import DataMixin

from pyxrd.generic.models import DataModel
from pyxrd.generic.io import storables, Storable
from pyxrd.generic.models.lines import StorableXYData

from pyxrd.file_parsers.json_parser import JSONParser

from pyxrd.calculations.goniometer import (
    get_lorentz_polarisation_factor,
    get_fixed_to_ads_correction_range,
    get_nm_from_2t, get_nm_from_t,
    get_2t_from_nm, get_t_from_nm,
)
from pyxrd.calculations.data_objects import GonioData 


@storables.register()
class Goniometer(DataModel, Storable):
    """
    The Goniometer class contains all the information related to the
    X-ray diffraction goniometer, e.g. wavelength, radius, slit sizes, ...
    """
    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Goniometer"

    _data_object = None
    @property
    def data_object(self):
        self._data_object.wavelength = self.wavelength
        x, y = self.wavelength_distribution.get_xy_data()
        self._data_object.wavelength_distribution = list(zip(x.tolist(), y.tolist()))
        return self._data_object

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: Start angle (in °2-theta, only  used when calculating without
    #: experimental data)
    min_2theta = FloatProperty(
        default=3.0, text="Start angle", minimum=0.0, maximum=180.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",  widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: End angle (in °2-theta, only  used when calculating without
    #: experimental data)
    max_2theta = FloatProperty(
        default=3.0, text="End angle", minimum=0.0, maximum=180.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",  widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The number of steps between start and end angle
    steps = IntegerProperty(
        default=2500, text="Steps", minimum=0, maximum=10000,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",  widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The wavelength distribution
    wavelength_distribution = LabeledProperty(
        default=None, text="Wavelength distribution",
        tabular=False, persistent=True, visible=True, 
        signal_name="data_changed", widget_type="xy_list_view",
        mix_with=(SignalMixin, ObserveMixin)
    )

    @FloatProperty(
        default=0.154056, text="Wavelength",
        tabular=True, persistent=False, visible=False,
        signal_name="data_changed",
        mix_with=(ReadOnlyMixin,)
    )
    def wavelength(self):
        """The wavelength of the generated X-rays (in nm)"""
        # Get the dominant wavelength in the distribution:
        x, y = self.wavelength_distribution.get_xy_data()
        wl = float(x[np.argmax(y)])
        return wl

    #: Flag indicating if the first soller slit is present or not
    has_soller1 = BoolProperty(
        default=True, text="Soller 1",
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The first Soller slit size (in °)
    soller1 = FloatProperty(
        default=2.3, text="Soller 1", minimum=0.0, maximum=10.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: Flag indicating if the second soller slit is present or not
    has_soller2 = BoolProperty(
        default=True, text="Soller 2",
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The second Soller slit size (in °)
    soller2 = FloatProperty(
        default=2.3, text="Soller 2", minimum=0.0, maximum=10.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The radius of the goniometer (in cm)
    radius = FloatProperty(
        default=24.0, text="Radius", minimum=0.0, maximum=200.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The divergence slit mode of the goniometer
    divergence_mode = StringChoiceProperty(
        default=settings.DEFAULT_DIVERGENCE_MODE, text="Divergence mode",
        visible=True, persistent=True, choices=settings.DIVERGENCE_MODES,
        signal_name="data_changed",
        mix_with=(DataMixin, SignalMixin,)
    )

    #: The divergence slit size (if fixed) or irradiated sample length (if automatic)
    divergence = FloatProperty(
        default=0.5, text="Divergence", minimum=0.0, maximum=90.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: Flag indicating if the second soller slit is present or not
    has_absorption_correction = BoolProperty(
        default=False, text="Absorption correction",
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The actual sample length
    sample_length = FloatProperty(
        default=1.25, text="Sample length [cm]", minimum=0.0,
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )
    
    #: The sample surface density
    sample_surf_density = FloatProperty(
        default=20.0, text="Sample surface density [mg/cm²]", minimum=0.0,
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )

    #: The sample mass absorption coefficient
    absorption = FloatProperty(
        default=45.0, text="Mass attenuation coeff. [cm²/g]", minimum=0.0,
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin)
    )
    
    #: Angular value (in degrees) for a monochromator correction - use 28.44 for silicon and 26.53 for carbon.
    mcr_2theta = FloatProperty(
        default=0.0, text="Monochromator 2θ", minimum=0.0, maximum=90.0,
        tabular=True, persistent=True, visible=True,
        signal_name="data_changed", widget_type="spin",
        mix_with=(DataMixin, SignalMixin,)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Constructor takes any of its properties as a keyword argument.
            
            In addition to the above, the constructor still supports the 
            following deprecated keywords, mapping to a current keyword:
                - lambda: maps to wavelength
                
            Any other arguments or keywords are passed to the base class.
        """
        my_kwargs = self.pop_kwargs(kwargs,
            "data_radius", "data_divergence", "data_soller1", "data_soller2",
            "data_min_2theta", "data_max_2theta", "data_lambda", "lambda",
            "wavelength", "has_ads", "ads_fact", "ads_phase_fact", "ads_phase_shift",
            "ads_const",
            *[prop.label for prop in Goniometer.Meta.get_local_persistent_properties()]
        )
        super(Goniometer, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self._data_object = GonioData()

        with self.data_changed.hold():
            self.radius = self.get_kwarg(kwargs, 24.0, "radius", "data_radius")
            
            #: Parse divergence mode (including old-style keywords):
            new_div_mode = self.get_kwarg(kwargs, None, "divergence_mode")
            if new_div_mode is None: # old style project
                old_ads = self.get_kwarg(kwargs, None, "has_ads")
                if old_ads is not None and old_ads: # if we have ads, set as such:
                    new_div_mode = "AUTOMATIC"
                else: # otherwise it was angular fixed slits
                    new_div_mode = settings.DEFAULT_DIVERGENCE_MODE
            self.divergence_mode = new_div_mode

            # Divergence value:                
            self.divergence = self.get_kwarg(kwargs, 0.5, "divergence", "data_divergence")

            # Monochromator correction:
            self.mcr_2theta = float(self.get_kwarg(kwargs, 0, "mcr_2theta"))

            # Soller slits:
            self.has_soller1 = self.get_kwarg(kwargs, True, "has_soller1")
            self.soller1 = float(self.get_kwarg(kwargs, 2.3, "soller1", "data_soller1"))
            self.has_soller2 = self.get_kwarg(kwargs, True, "has_soller2")
            self.soller2 = float(self.get_kwarg(kwargs, 2.3, "soller2", "data_soller2"))

            # Angular range settings for calculated patterns:
            self.min_2theta = float(self.get_kwarg(kwargs, 3.0, "min_2theta", "data_min_2theta"))
            self.max_2theta = float(self.get_kwarg(kwargs, 45.0, "max_2theta", "data_max_2theta"))
            self.steps = int(self.get_kwarg(kwargs, 2500, "steps"))

            # Sample characteristics
            self.sample_length = float(self.get_kwarg(kwargs, settings.DEFAULT_SAMPLE_LENGTH, "sample_length"))
            self.absorption = float(self.get_kwarg(kwargs, 45.0, "absorption"))
            self.sample_surf_density = float(self.get_kwarg(kwargs, 20.0, "sample_surf_density"))
            self.has_absorption_correction = bool(self.get_kwarg(kwargs, False, "has_absorption_correction"))
                       
            wavelength = self.get_kwarg(kwargs, None, "wavelength", "data_lambda", "lambda")
            if not "wavelength_distribution" in kwargs and wavelength is not None:
                default_wld = [ [wavelength,1.0], ]
            else:
                # A Cu wld:
                default_wld = [
                    [0.1544426,0.955148885],
                    [0.153475,0.044851115],
                ]                
            self.wavelength_distribution = StorableXYData(
               data=self.get_kwarg(kwargs, list(zip(*default_wld)), "wavelength_distribution")
            )           

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def __reduce__(self):
        return (type(self), ((), self.json_properties()))

    def json_properties(self):
        props = Storable.json_properties(self)
        props["wavelength_distribution"] = self.wavelength_distribution._serialize_data()
        return props

    def reset_from_file(self, gonfile):
        """
        Loads & sets the parameters from the goniometer JSON file
        specified by `gonfile`, can be a filename or a file-like object.
        """
        new_gonio = JSONParser.parse(gonfile)
        with self.data_changed.hold():
            for prop in self.Meta.all_properties:
                if prop.persistent:
                    if prop.label == "wavelength_distribution":
                        self.wavelength_distribution.clear()
                        self.wavelength_distribution.set_data(
                            *new_gonio.wavelength_distribution.get_xy_data())  
                    elif prop.label != "uuid":
                        setattr(self, prop.label, getattr(new_gonio, prop.label))
                        
    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_nm_from_t(self, theta):
        """Converts a theta position to a nanometer value"""
        return get_nm_from_t(
            theta,
            wavelength=self.wavelength, zero_for_inf=True
        )

    def get_nm_from_2t(self, twotheta):
        """Converts a 2-theta position to a nanometer value"""
        return get_nm_from_2t(
            twotheta,
            wavelength=self.wavelength, zero_for_inf=True
        )

    def get_t_from_nm(self, nm):
        """ Converts a nanometer value to a theta position"""
        return get_t_from_nm(nm, wavelength=self.wavelength)

    def get_2t_from_nm(self, nm):
        """ Converts a nanometer value to a 2-theta position"""
        return get_2t_from_nm(nm, wavelength=self.wavelength)

    def get_default_theta_range(self, as_radians=True):
        """
        Returns a numpy array containing the theta values as radians from
        `min_2theta` to `max_2theta` with `steps` controlling the interval.
        When `as_radians` is set to False the values are returned as degrees. 
        """
        def torad(val):
            if as_radians:
                return radians(val)
            else:
                return val
        min_theta = torad(self.min_2theta * 0.5)
        max_theta = torad(self.max_2theta * 0.5)
        delta_theta = float(max_theta - min_theta) / float(self.steps)
        theta_range = (min_theta + delta_theta * np.arange(0, self.steps, dtype=float)) + delta_theta * 0.5
        return theta_range

    def get_lorentz_polarisation_factor(self, range_theta, sigma_star):
        """
            Calculates Lorentz polarization factor for the given theta range
            and sigma-star value using the information about the goniometer's
            geometry.
        """
        return get_lorentz_polarisation_factor(
            range_theta, sigma_star, self.soller1, self.soller2, self.mcr_2theta
        )

    def get_ADS_to_fixed_correction(self, range_theta):
        """
            Returns a correction range that will convert ADS data to fixed slit
            data. Use with caution.
        """
        return 1.0 / get_fixed_to_ads_correction_range(range_theta, self.data_object)

    pass # end of class
